% LICENSE AGREEMENT
% - - - - - - - - -
% Copyright (c) 2022, Stilianos Louca
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
%
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in
%      the documentation and/or other materials provided with the distribution
%    * Neither the name of the original author (Stilianos Louca), nor the names
%      of any other contributors may be used to endorse or promote products derived
%      from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%
% - - - - - - - - - - - - - - - - - - - - - -
%
% This MATLAB code demonstrates fast-reaction-transport modeling in an aquatic laterally homogeneous sediment column.
% We consider two microbially catalyzed coupled redox reactions: sulfate-dependent anaerobic oxidation of methane (CH4 + SO4 + 2H -> H2S + CO2 + 2H_2O)
% and aerobic oxidation of sulfide (H2S + 2O2 -> SO4 + 2H), with methane originating from deeper sediment layers, oxygen originating from the water column,
% and H2S and SO4 acting as temporary intermediates ("electron shuttles") between the two reactions. We assume that transport occurs exclusively via molecular diffusion,
% with diffusivities for each substance taken from the literature. We are interested in the system's response to seasonal variations in oxygen concentrations
% at the water-sediment interface, and hence we set the O2 concentration at the top boundary to a sinusoidally oscillating function of time, 
% while keeping the boundary concentrations of all other substances constant. 
% As initial condition we assume a recently mixed and aerated sediment column (i.e., a constant non-zero O2 and SO4 profile between the top and bottom boundary, with methane and sulfide absent).
% 
% Time is measured in days, depth in meters, metabolite concentrations in mol/L
% To execute this code, simply run:
%	demo_sulfate_methane_transition()
% Note that the code makes use of various external MATLAB code files (most notably "runFRTmodel.m" and "plotFRTresults.m"), which should be placed in the MATLAB path.
% Tested on MATLAB R2022a
%
% Stilianos Louca
% March 04, 2022
%
function demo_sulfate_methane_transition()
	
	% Close any open figures (save memory)
	close all hidden;
	
	% display more significant digits when printing variables to the screen
	format long;
		
	% figure out uniquely numbered output dir
	outputDir = getNonExistingSimilarDir('output', 'sulfate_methane_');
	makedir(outputDir);
	
	% prepare console output log
	diary(strcat(outputDir,'/console_output.txt'));
	
	fprintf('Initializing..\nNote: All outputs and plots will be saved in the directory "%s"\n', outputDir);
	
	% make backup copies of code
	fprintf('Making backup copies of source code..\n');
	makedir(strcat(outputDir,'/code_backup'));
	sources = dir('*.m');
	for s=1:length(sources)
		copyfile(sources(s).name, strcat(outputDir, '/code_backup/', sources(s).name)); 
	end

	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% Define model structure and other options
	
	fprintf('Defining model and other options..\n');

	RANDOM_SEED = 123;

	% time (days) & depth (meters) grid
	days_per_year = 365.25;
	TIMES 	= linspace(0,4*days_per_year,600);
	DEPTHS	= linspace(0,0.2,200);

	% define pool of metabolites & reactions
	% note that the actual FRT model may focus on a subset of metabolites & reactions, specified further below
	metabolites = {'H2S'; 'O2'; 'SO4'; 'H'; 'CH4'; 'CO2'; 'H2O'};
	reactions	= {'oxH2StoSO4_wO2'; 'oxCH4toCO2_rSO4toH2S'};
	
	DeltaG0_ann	= {	'oxH2StoSO4_wO2' 		-751.8;
					'oxCH4toCO2_rSO4toH2S'	-107.9};
					
	% acceptable errors for metabolite concentrations when solving the fast-reaction ODE
	ODE_concentration_errors = [0.01e-6;	% H2S
								0.01e-6;	% O2
								0.01e-6;	% SO4
								0.01e-6;	% H
								0.01e-6;	% CH4
								0.01e-6;	% CO2
								1e-6];		% H2O
								
	
	% Stoichiometric matrix (size M*R)
	%				Reaction				Metabolite		Stoich.coeff.		% limiting (kinetics)
	stoich_ann	= {	'oxH2StoSO4_wO2'		'H2S'			-1					true;
					'oxH2StoSO4_wO2'		'O2'			-2					true;
					'oxH2StoSO4_wO2'		'SO4'			1					false;
					'oxH2StoSO4_wO2'		'H'				2					false;
					'oxCH4toCO2_rSO4toH2S'	'CH4'			-1					true;
					'oxCH4toCO2_rSO4toH2S'	'SO4'			-1					true;
					'oxCH4toCO2_rSO4toH2S'	'H'				-2					false;
					'oxCH4toCO2_rSO4toH2S'	'CO2'			1					false;
					'oxCH4toCO2_rSO4toH2S'	'H2S'			1					false;
					'oxCH4toCO2_rSO4toH2S'	'H2O'			2					false};
	
	% define initial concentrations of metabolites as functions of depth
	initial_concentrations = {	@(z) zeros(size(z));			% H2S
								@(z) ones(size(z)) * 160e-6;	% O2
								@(z) ones(size(z)) * 300e-6;	% SO4. Typical SO4 concentration in freswater lakes (e.g., Mostovaya et al. 2021)
								@(z) ones(size(z)) * 2.399e-8;	% [H] = 2.399e-08 M, based on pH=7.62
								@(z) zeros(size(z));			% CH4
								@(z) ones(size(z)) * 0.027e-6;	% CO2. Mean CO2 concentration at bottom of Cariaco Basin.
								@(z) ones(size(z))};			% H2O
								
	% define top boundary conditions of metabolites as functions of time
	% 					BC_type		BC_value
	top_conditions = {	0			@(t) zeros(size(t));			% H2S
						0			@(t) 160e-6 * (1 + 0.75 * sin((2*pi/days_per_year) * t));	% O2
						0			@(t) ones(size(t)) * 300e-6;	% SO4. Typical SO4 concentration in freshwater lakes (e.g., Mostovaya et al. 2021)
						0			@(t) ones(size(t)) * 2.399e-8;	% [H] = 2.399e-08 M, based on pH=7.62 (Cariaco at 300 m)
						0			@(t) zeros(size(t));			% CH4
						0			@(t) ones(size(t)) * 0.027e-6;	% CO2. Mean CO2 concentration at bottom of Cariaco Basin.
						0			@(t) ones(size(t))};			% H2O

	% define bottom boundary conditions of metabolites as functions of time
	% 						BC_type		BC_value
	bottom_conditions = {	1			@(t) zeros(size(t));			% H2S
							0			@(t) zeros(size(t));			% O2
							0			@(t) zeros(size(t));			% SO4
							0			@(t) ones(size(t)) * 2.399e-8;	% [H] = 2.399e-08 M, based on pH=7.62 (Cariaco at 300 m)
							0			@(t) ones(size(t)) * 300e-6;	% CH4.
							0			@(t) ones(size(t)) * 0.027e-6;	% CO2. Mean CO2 concentration at bottom of Cariaco Basin.
							0			@(t) ones(size(t))};			% H2O
								
	% define temperature (in Celsius) as a vectorized function of time and depth
	temperature = @(t,z) ones(size(t)) * 18;
	
	% define diffusivities (in m^2/day) as a vectorized function of time and depth, for each metabolite
	diffusivities =	{@(t,z) ones(size(t)) * 8.2e-5;	% H2S. Based on a HS^-/SO4 self-diffusion coefficient ratio of 1.7 (Roden and Tuttle, 1992, Limnology and Oceanography. 37:725-738) and a sulfate diffusivity of 4.8e-5 (Niewöhner et al. 1998).
					 @(t,z) ones(size(t)) * 9.4e-5;	% O2. Based on [Han and Bartels (1996). Temperature dependence of Oxygen diffusion in H2O and D2O. 100:5597-5602. Table 1]
					 @(t,z) ones(size(t)) * 4.8e-5;	% SO4. Based on [Niewöhner et al. (1998). Deep sulfate reduction completely mediated by anaerobic methane oxidation in sediments of the upwelling area off Namibia. Geochimica et Cosmochimica Acta. 62:455-464. Fig. 5]
					 @(t,z) NaN(size(t));			% H
					 @(t,z) ones(size(t)) * 7.5e-5;	% CH4. Based on [Niewöhner et al. (1998). Deep sulfate reduction completely mediated by anaerobic methane oxidation in sediments of the upwelling area off Namibia. Geochimica et Cosmochimica Acta. 62:455-464. Fig. 5]
					 @(t,z) NaN(size(t));			% CO2
					 @(t,z) NaN(size(t))};			% H2O
	
		
	% define vertical downward advection speed (in m/day) as a vectorized function of time and depth
	advection = @(t,z) zeros(size(t));
	
	% define the relative slope of the lateral (cross-sectional) area as a function of depth.
	% This is (1/A)*dA/dz, where A is the lateral area (e.g. of a modeled ocean basin). 
	% This is used to correct for geometric dilution effects, in case the lateral area varies with depth. If you don't know what this is, you can have this function simply return 0.
	dilution_slope = @(t,z) zeros(size(t));
	
	% specify subset of reactions & metabolites to consider for FRT modeling
	FRT_reactions 	= {'oxH2StoSO4_wO2'; 'oxCH4toCO2_rSO4toH2S'};
	FRT_metabolites = { 'H2S'; 'O2'; 'SO4'; 'CH4' };

				
	FRT_INTERLEAVE_TIMESTEP 	= 0.2; 		% time interval between the pure transport model and the reaction part (in days). A smaller value implies a greater consistency with the FRT limit (and in particular sharper reaction fronts) at the expense of longer computation time. This parameter has a strong effect on the computational efficiency of the FRT algorithm.
	FRT_TRANSPORT_TIMESTEP 		= 0.1; 		% suggested integration timestep for FRT transport part (in days).
	FRT_TRANSPORT_REL_ERROR 	= 0.005; 	% relative error for PDE solver for transport part
	FRT_REACTION_REL_ERROR 		= 0.005;	% relative error for ODE solver for reaction part
	FRT_PARALLEL 				= true;		% use parallel computing where possible
	FRT_DELTAG_THRESHOLD		= 0; 		% kJ/mol
	FRT_REACTIONS_MAX_RUNTIME	= 10;		% maximum allowed runtime (wall time in seconds) per fast reaction stage per grid cell. If NaN, this stopping criterion is ignored.
	FRT_REVERSIBLE				= false;	% assume that all FRT reactions are reversible

	% (mol/l) smallness threshold for metabolite concentrations
	% To prevent zero or Inf reaction quotients, make sure this is much smaller than realistic metabolite concentrations.
	C_EPSILON = 1e-14;

	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% Parse model definition

	fprintf('Parsing model definition and other options..\n');
	
	% seed random number generator
	if(RANDOM_SEED<=0)
		% use current time to seed
		rng('shuffle');
		s = rng;
		fprintf('  Seeding random number generator with random seed %d..\n',s.Seed);
	else
		% use fixed (provided) seed
		fprintf('  Seeding random number generator with fixed seed %d..\n',RANDOM_SEED);
		rng(RANDOM_SEED);
	end

	M  = length(metabolites);
	R  = length(reactions);
	
	% create stoichiometric matrix and matrix of limiting metabolites for each reaction
	[stoich, limiting, errormsg] = deannotateAndUncompressStoich(stoich_ann, metabolites, reactions);
	if(not(strcmp(errormsg,'')))
		fprintf('  Error in stoich_ann: %s\n', errormsg);
		diary off; return;
	end
		
	% Standard Gibbs energy of redox reactions
	DeltaG0 = arrayfun(@(r) DeltaG0_ann{find(strcmp(DeltaG0_ann(:,1),reactions{r}),1),2}, 1:R);
	
	% sub-reaction network to be considered for FRT
	[FRT_sub_stoich, FRT_sub_reactions_indices, FRT_sub_metabolites_indices, errormsg] = deannotateSubReactionNetwork(FRT_reactions, FRT_metabolites, stoich, reactions, metabolites, '');
	if(not(strcmp(errormsg,'')))
		fprintf('  Error in FRT_reactions{} or FRT_metabolites{}: %s\n', errormsg);
		diary off;
		return;
	else
		fprintf('  Note: FRT sub-reaction network comprises %d reactions and %d metabolites\n',length(FRT_sub_reactions_indices),length(FRT_sub_metabolites_indices));
	end
	sub_metabolites = metabolites(FRT_sub_metabolites_indices);
	sub_reactions	= reactions(FRT_sub_reactions_indices);
	Msub 			= length(FRT_sub_metabolites_indices);
	Rsub 			= length(FRT_sub_reactions_indices);

	
	% define effective reaction equilibrium constants as a single function of time & depth
	% Note that we may be considering ("tracking") only a subset of metabolites actually involved in the reactions, so our equilibrium constants should effectively refer to only those tracked metabolites.
	% We thus define the effective equilibrium constant, Keffective, as the value of \prod_{m considered} C_m^{stoich_{mr}} at which the reaction has zero DeltaG.
	% Hence, we must correct our equilibrium constants to account for the fixed concentrations of "background" metabolites: ln(Keffective) = ln(Ktrue) - \sum_{m background} stoich_{mr} * ln(C_m)
	% Here we assume that the concentrations of background metabolites are constant over time, and thus equal to their initial concentration profiles
	Rg = 8.3144621e-3; % gas constant (kJ/mol/K)
	omitted_metabolite_indices = get_complement(M, FRT_sub_metabolites_indices);
	omitted_stoich = stoich(omitted_metabolite_indices,FRT_sub_reactions_indices);
	function lK = logK(t,z)
		omitted_concentrations = arrayfun(@(m) initial_concentrations{m}(z), omitted_metabolite_indices);
		lK = -(DeltaG0(FRT_sub_reactions_indices)./(Rg * (273.15+temperature(t,z))));
		lK = lK - reshape(omitted_stoich.' * log(maxNan(C_EPSILON,omitted_concentrations)), size(lK));
	end


	fprintf('Simulating model over times %g-%g and depths %g-%g m, %d reactions x %d metabolites..\n',TIMES(1),TIMES(end),DEPTHS(1),DEPTHS(end),Rsub,Msub);
	[concentrations,...
	net_sources,...
	reaction_rates,...
	energy_rates,...
	deltaLogK,...
	ndeltaLogK,...
	tendencies,...
	errormsg] = runFRTmodel(TIMES,...
							DEPTHS,...
							Msub, ...
							Rsub, ...
							sub_metabolites,...
							sub_reactions,...
							cell2mat(top_conditions(FRT_sub_metabolites_indices,1)),...
							cell2mat(bottom_conditions(FRT_sub_metabolites_indices,1)),...
							top_conditions(FRT_sub_metabolites_indices,2), ...
							bottom_conditions(FRT_sub_metabolites_indices,2), ...
							initial_concentrations(FRT_sub_metabolites_indices), ...
							diffusivities(FRT_sub_metabolites_indices),...
							{},...	% since we're using Fick's law, here's no need to provide a diffusivity gradient
							{advection},...
							dilution_slope,...
							temperature,...
							true,...
							FRT_sub_stoich, ...
							@logK,...
							ternary(FRT_REVERSIBLE,true(Rsub,1),false(Rsub,1)),... % specify reversible reactions
							true,...					% check_nans_in_input
							FRT_INTERLEAVE_TIMESTEP,... % interleave_dt
							FRT_TRANSPORT_TIMESTEP,...  % transport_dt
							FRT_TRANSPORT_REL_ERROR,... % transport_relative_error
							FRT_REACTION_REL_ERROR,...  % reaction_relative_errors
							ODE_concentration_errors(FRT_sub_metabolites_indices),...		% concentration_errors
							FRT_DELTAG_THRESHOLD,...
							NaN,...				% use thermodynamic force as kinetics
							FRT_REACTIONS_MAX_RUNTIME,...
							FRT_PARALLEL,...
							true,...			% diagnostics
							true,...			% verbose
							'  '); 				% verbose_prefix
	if(not(strcmp(errormsg,'')))
		fprintf('  ERROR: Could not simulate model: %s\n',errormsg);
	else
		fprintf('Plotting FRT results..\n');
		plotFRTresults(	TIMES,...
						DEPTHS,...
						sub_metabolites,...
						sub_reactions,...
						concentrations,...
						net_sources,...
						reaction_rates,...
						energy_rates,...
						deltaLogK,...
						ndeltaLogK,...
						tendencies,...
						sprintf('%s/FRT',outputDir),...
						'  ');
	end
	
	fprintf('\nFinished. All outputs are in: %s\n',outputDir);
	diary off;
end



% ************************************* %
% AUXILIARY FUNCTIONS
% ************************************* %
							


% create a directory, if it does not exit
function makedir(path)
	if ~exist(path, 'dir')
		mkdir(path);
	end
end



% Get file path to a directory for which there exists no similarly named directory or file
% dir names are chosen incrementally, basename001, basename002 and so on, until no basenameXXX* exists
function filepath = getNonExistingSimilarDir(parent, basename)
	% get all items (level 1) in parent dir
	listing  = dir(sprintf('%s/%s*',parent,basename));
	existing = cell(length(listing),1);
	for l=1:length(listing)
		existing{l} = listing(l).name;
	end
	% increment counter until no similarly named item exists
	n = 1;
	while(true)
		filename = sprintf('%s%03d',basename,n);		
		if(not(any(strncmpi(existing,filename,length(filename)))))
			break;
		end
		n = n+1;
	end
	filepath = sprintf('%s/%s',parent,filename);
end



% Takes annotated compressed stoichiometric matrix (stoich_ann) and returns de-annotated and uncompressed (expanded) version of it
function [S, limiting, errormsg] = deannotateAndUncompressStoich(stoich_ann, metabolites, reactions)
	rows 	= length(metabolites);
	cols	= length(reactions);
	S	 	= zeros(rows,cols);
	limiting= false(rows,cols);
	for i=1:size(stoich_ann,1)
		[valid, col] = ismember(stoich_ann{i,1}, reactions);
		if(not(valid))
			errormsg = ['Gene name "' stoich_ann{i,1} '" not recognized'];
			return;
		end
		[valid, row] = ismember(stoich_ann{i,2}, metabolites);
		if(not(valid))
			errormsg = ['Metabolite name "' stoich_ann{i,2} '" not recognized'];
			return;
		else
			S(row,col) = stoich_ann{i,3};
		end
		limiting(row,col) = stoich_ann{i,4};
	end
	errormsg='';
end



% returns a sub-stoichiometric matrix constrained to a subset of reactions (reactions) and metabolites
% sub_reactions_ann{} and sub_metabolites_ann{} should contain the names of the considered reactions and metabolites
% the full stoichiometric matrix, gene names and metabolite names must be given via stoich(,), reactions{} and metabolites{}
% if everythings is OK then errormsg will be ''
function [sub_stoich, sub_reactions_indices, sub_metabolites_indices, errormsg] = deannotateSubReactionNetwork(sub_reactions_ann, sub_metabolites_ann, stoich, reactions, metabolites, verbose_prefix)
	RS = length(sub_reactions_ann);
	MS = length(sub_metabolites_ann);
	
	% default return values
	sub_reactions_indices 		= NaN;
	errormsg 				= '';
	sub_metabolites_indices = NaN;
	sub_stoich 				= NaN;
	
	%if(RS>MS) fprintf('%sWARNING: Number of reactions (%d) in sub-reaction network exceeds number of metabolites (%d)\n',verbose_prefix,RS,MS); end
	
	sub_reactions_indices = zeros(RS, 1);
	for rs=1:RS
		[valid, r] = ismember(sub_reactions_ann{rs}, reactions);
		if(not(valid))
			errormsg = ['Unknown gene name "' sub_reactions_ann{rs} '" in sub network'];
			return;
		else
			sub_reactions_indices(rs) = r;
		end
	end
	
	sub_metabolites_indices = zeros(MS, 1);
	for ms=1:MS
		[valid, m] = ismember(sub_metabolites_ann{ms}, metabolites);
		if(not(valid))
			errormsg = ['Unknown metabolite name "' sub_metabolites_ann{ms} '" in sub network'];
			return;
		else
			sub_metabolites_indices(ms) = m;
		end
	end
	
	sub_stoich = zeros(MS, RS);
	for rs=1:RS
		sub_stoich(:,rs) = stoich(sub_metabolites_indices,sub_reactions_indices(rs));
	end
end


% calculate max(a,b,dim) while propagating nans
% that is, when a comparison with NaN is performed, the result is NaN
function [maxvalues, indices] = maxNan(A, B, dim)
	if(nargin==2)
		if(nargout == 2)
			% max throws error. No need to worry about NaNs.
			[maxvalues, indices] = max(A, B);
		else
			maxvalues = max(A, B);
			maxvalues(isnan(A) | isnan(B)) = NaN;
		end
	else	
		if(nargout==2)
			if(nargin==3)
				[maxvalues, indices] = max(A, B, dim);
				nans = any(isnan(A), dim);
			else
				[maxvalues, indices] = max(A);
				nans = any(isnan(A));
			end
			maxvalues(nans) = NaN;
			indices(nans) = NaN;
		else
			if(nargin==3)
				maxvalues = max(A, B, dim);
				maxvalues(any(isnan(A), dim)) = NaN;
			else
				maxvalues = max(A);
				maxvalues(any(isnan(A))) = NaN;
			end
		end
	end
end

function complement = get_complement(N, subset)
	keep = true(N,1);
	keep(subset) = false;
	complement = find(keep);
end



% recursively gzip all files in a given folder, removing the original files
function compress_dir_contents(dirpath)
	if(exist(dirpath, 'dir'))
		compressed_files = gzip(dirpath);
		for f=1:length(compressed_files)
			compressed_file = compressed_files{f};
			uncompressed_file = compressed_file(1:end-3);
			if(exist(uncompressed_file, 'file'))
				delete(uncompressed_file);
			end
		end
	end
end


% Ternary operator for C-like conditional expressions
% Note: Calling this will evaluate both a and b, despite only one being returned
function c = ternary(condition, a, b)
     try
		if condition
			c = a;
		else
			c = b;
		end
	catch
		c = false;
	end
end
