% LICENSE AGREEMENT
% - - - - - - - - -
% Copyright (c) 2022, Stilianos Louca
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
%
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in
%      the documentation and/or other materials provided with the distribution
%    * Neither the name of the original author (Stilianos Louca), nor the names
%      of any other contributors may be used to endorse or promote products derived
%      from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%
% - - - - - - - - -
% Tested on MATLAB R2019b.
% - - - - - - - - -
%
% DESCRIPTION
% - - - - - -
% Plot spatiotemporal dynamics for N observables as contour plots
% values(ti,zi,j) is value of j-th observable at time times(ti) and depth depths(zi).
% titles{} is either a cell array of N strings or a single string. Similarly for cblabels{} (for the color-bars)
function plotSpatiotemporal(times, ...
							depths, ...
							values, ...
							scatterPoints, ... 	% cell array of 2D scatterpoints as 2-column arrays, where dataPoints{j}(i,1:2) is one scatterpoint in the j-th plot. Note that dataPoints{j} can have more than 2 columns, however these will be ignored. Scatterpoints will appear as black dots. These could represent, for example, the original data points used in an interpolation. scatterPoints{} can also be empty or of size 1 (in the latter case the same scatterpoints will be shown in all plots).
							rescalings,...	 	% 3-ple specifying (time,depth,value) rescaling factors. Can also be an empty tuple (no rescaling)
							names,...			% 1D cell array of strings of length N, specifying brief names for the various plots. Used e.g. to generate file names. For example {'NH4', 'O2', 'SO4'}.
							titles, ...
							timeLabels, ...		% 1D cell array of strings or a single string, e.g. 'year'
							depthLabel,...		% e.g. 'depth (m)'
							cblabels, ...
							logarithmicColorMap,...		% either a 1D cell array of N booleans or a single boolean
							min_cbvalues, ...			% 1D numerical array of size N, or a single value. Can also contain NaNs. Minimum value on colorbar. If logarithmicColorMap==true, this must be NaN or >0.
							plotFile, ...
							supertitle, ...
							plotWidth, ...
							plotHeight, ...
							numberOfContours, ...
							equalizeColorRange, ... 	% if true, only the 2% - 98% centile range (if logarithmicColorMap) or 0%-98% centile range (if not logarithmicColorMap) of the values is linearly color-mapped (the rest is squeezed at the colorbar edges). This is to avoid a few outliers from screwing up the color map
							dataFile,... % either a path to a single text file, or a basepath to multiple text files (depending on split_dataFile), for saving scatterdata. Can be '' (empty).
							split_dataFile)	% if true, then the data will be written in separate files, one file per plot
	N  = size(values,3);
	NZ = size(values,2);
	NT = size(values,1);
	
	% geometry for sub-plot distribution
	[nrows ncols] = num2RowsColumns(N, 0.6,0.2,3);
	titleAreaV	= 0.05 * (1+length(strfind(supertitle,sprintf('\n')))); % reserve title space
	plotAreaV 	= 1-titleAreaV;
	
	% initialize figure window (make invisible since we plot everything to pdf)
	fig = figure('Visible','off','PaperPosition',[0 0 ncols*plotWidth nrows*plotHeight],'PaperSize',[ncols*plotWidth nrows*plotHeight]);

	% apply rescalings
	if(length(rescalings)>=3)
		times = times*rescalings(1);
		depths = depths*rescalings(2);
		values = values*rescalings(3);
	end

	% re-format input if needed
	if(~iscell(names))
		names = repmat({names},N,1);
	end
	if(~iscell(titles))
		titles = repmat({titles},N,1);
	end
	if(~iscell(cblabels))
		cblabels = repmat({cblabels},N,1);
	end
	if(not(isempty(dataFile)) & split_dataFile && (endsWith(dataFile,'.tsv') || endsWith(dataFile,'.txt')))
		% user provided a TSV file path even though they should have provided a basepath, so convert it into a basepath
		dataFile = dataFile(1:end-3);
	end
	
	% iterate through sub plots
	for row=1:nrows
		for col=1:ncols
			j = (row-1)*ncols+col;
			if(j>N) break; end
			
			if(iscell(logarithmicColorMap))
				logarithmic = logarithmicColorMap{j};
			else
				logarithmic = logarithmicColorMap;
			end
			if(iscell(timeLabels))
				timeLabel = timeLabels{j};
			else
				timeLabel = timeLabels;
			end
			if(length(min_cbvalues)>1)
				min_cbvalue = min_cbvalues(j);
			else
				min_cbvalue = min_cbvalues(1);
			end

			% mannually place subplot
			% Note: If subplot overlaps previous subplots, the older ones will be deleted, unless overlap is exact! (this is an idiotic matlab 'feature')
			subplot_position = [(0.15+col-1)/ncols plotAreaV*(0.25+nrows-row)/nrows 0.7/ncols plotAreaV*0.6/nrows];
			subplot('Position',subplot_position);
			
			values_to_plot = squeeze(values(:,:,j));
			all_nan = all(all(isnan(values_to_plot) | isinf(values_to_plot)));
			if(all_nan)
				% all values are NaN, so just show an error message as plot title
				title(sprintf('%s: All NaN or Inf',sanitize_string_for_latex(titles{j})),'fontsize',9);
				continue;
			end
			if(equalizeColorRange)
				centiles = prctile(reshape(values_to_plot,[1 length(times)*length(depths)]), [2 98]);
				if(logarithmic & (centiles(1)<centiles(2)) & (centiles(1)>0) & not(any(isnan(centiles))))
					% make sure values below the lower centile are not zero. This "hack" will not affect the plot since they are out of the shown range
					below_centiles = (values_to_plot<centiles(1));
					values_to_plot(below_centiles) = centiles(1)/2;
				else
					% modify outlier values prior to even plotting, so that contour lines and color assignments are adjusted to the actually plotted value range
					values_to_plot(values_to_plot<centiles(1)) = centiles(1);
					values_to_plot(values_to_plot>centiles(2)) = centiles(2);
				end
				if(any(isnan(centiles)))
					warning('off','MATLAB:contourf:NonFiniteData'); % plotting will fail for this subplot, but turn this annoying warning off
				end
			elseif(logarithmic && any(any(values_to_plot<=0)))
				% replace zeros and negatives with a small value
				small_value = min(values_to_plot(values_to_plot>0));
				if(isnan(small_value))
					logarithmic = false;
				else
					small_value = small_value/10;
				end
				values_to_plot(values_to_plot<=0) = small_value;
			end
		
			if(logarithmic) 
				values_to_plot = log(values_to_plot);
				if(equalizeColorRange) centiles = log(centiles); end
			end
			values_to_plot(isinf(values_to_plot)) = NaN; % replace inf with NaN, because contourf doesn't like infinities

			if(not(isnan(min_cbvalue)))
				values_to_plot(values_to_plot<min_cbvalue) = min_cbvalue; % replace out-of-bounds-values with min_cbvalue for more even contour grading
			end

		
			% create contour-plot
			% turn off some annoying warning while doing so
			warning('off','MATLAB:contourf:ConstantData');
			warning('off','MATLAB:contour:ConstantData');
			warning('off','MATLAB:contour:NonFiniteData');
			try
				contourf(times,depths,values_to_plot', numberOfContours,'LineColor','none');
			catch exception
				fprintf('ERROR during contour plot\n');
				return;
			end
			shading flat;

			% rescale scatterpoint coordinates
			if((length(scatterPoints)>=j) & (size(scatterPoints{j},1)>1) & (length(rescalings)>=3))
				scatterPoints{j}(:,1) = scatterPoints{j}(:,1)*rescalings(1);
				scatterPoints{j}(:,2) = scatterPoints{j}(:,2)*rescalings(2);
			end
					
			% overlay scatterpoints
			if((length(scatterPoints)==1) & (size(scatterPoints{1},1)>1))
				hold on
				scatter(scatterPoints{1}(:,1), scatterPoints{1}(:,2), 0.6, 'black','+');
				hold off
			elseif((length(scatterPoints)>=j) & (size(scatterPoints{j},1)>1))
				hold on
				scatter(scatterPoints{j}(:,1), scatterPoints{j}(:,2), 0.6, 'black','+');
				hold off
			end
		
			% axes and titles
			set(gca,'YDir','reverse');
			title(sanitize_string_for_latex(titles{j}),'fontsize',9, 'Interpreter','latex');
			ylabel(depthLabel,'fontsize',9);
			xlabel(timeLabel,'fontsize',9);
		
			cb = colorbar;
		
			% decrease colorbar width and re-position
			temp = get(gca,'Position');
			cpos = get(cb,'Position');
			cpos(3) = (0.65*cpos(3))/ncols;
			cpos(1) = subplot_position(1) + 1.02 * subplot_position(3); % fix x-position of color bar, which when extracted using get() is not consistent with the original subplot_position
			set(cb,'Position',cpos);
			set(gca,'Position',temp);
		
			% colorbar label
			ylabel(cb, sanitize_string_for_latex(cblabels{j}), 'Interpreter','latex');
		
			set(gca,'TickDir','out')
			set(gca,'fontsize',8)

			% restrict color range to within centiles
			if(equalizeColorRange & not(any(isinf(centiles))) & not(any(isnan(centiles))) & (centiles(1)<centiles(2)))
				if(isnan(min_cbvalue))
					min_cbvalue = centiles(1);
				end
				if(centiles(2)>min_cbvalue)
					max_cbvalue = centiles(2);
				else
					max_cbvalue = max(max(values_to_plot));
				end
			else
				if(isnan(min_cbvalue))
					min_cbvalue = min(min(values_to_plot));				
				end
				max_cbvalue = max(max(values_to_plot));
			end
			if(min_cbvalue>=max_cbvalue)
				continue; % can't plot with trivial color range
			end
			caxis([min_cbvalue, max_cbvalue]);
		
			% correct ticks if logarithmic
			if(logarithmic)
				original_ticks = get(cb, 'YTick');
				set(cb,'YTick',original_ticks,'YTickLabel',exp(original_ticks));
			end
			
			% specify color map
			%colormap(colormap_red_white_blue(numberOfContours+1));
			colormap(jet(numberOfContours+1));
		end
	end
	
	%create super title
	%this should come after subplot(), because the latter has a bad tendency to delete previous figure elements
	annotation('textbox', [0 (1-titleAreaV) 1 titleAreaV], 'String', sanitize_string_for_latex(supertitle), 'EdgeColor', 'none', 'HorizontalAlignment', 'center', 'VerticalAlignment', 'top','Interpreter','latex');

	warning('off','MATLAB:hg:patch:PatchFaceVertexCDataLengthMustEqualVerticesLength'); % turn this annoying warning off
	make_parent(plotFile);
	print(fig,'-dpdf',plotFile);
	
	% Write data to file
	if(not(isempty(dataFile)))
		supertitle_sanitized = strrep(supertitle, sprintf('\n'), sprintf('\n#   '));
		if(split_dataFile)
			% write each dataset into a separate file (one file per plot)
			for j=1:N
				name = names{j};
				filepath = strcat(dataFile,name,'.contour.tsv');
				make_parent(filepath);
				fout = fopen(filepath, 'w');
				fprintf(fout, '# %s\n# Spatiotemporal contour data for: %s, %s\n# time_depth\ttime\tdepth\t%s\n', supertitle_sanitized, titles{j}, cblabels{j}, name);
				for ti=1:NT
					for zi=1:NZ
						fprintf(fout, '%g_%g\t%g\t%g\t%g\n', times(ti), depths(zi), times(ti), depths(zi), values(ti,zi,j)); 
					end
					fprintf(fout, '\n');
				end
				fclose(fout);
				% also create a TSV file for any scatterdata
				if((length(scatterPoints)==1) || (length(scatterPoints)>=j))
					if(length(scatterPoints)==1)
						XY= scatterPoints{1};
					elseif(length(scatterPoints)>=j)
						XY= scatterPoints{j};
					end
					fout = fopen(strcat(dataFile,name,'.scatter.tsv'), 'w');
					fprintf(fout, '# %s\n# Spatiotemporal scatterpoints for: %s, %s\n# time\tdepth\n', supertitle_sanitized, titles{j}, name);
					for p=1:length(XY)
						fprintf(fout, '%g\t%g\n', XY(p,1), XY(p,2));
					end
					fclose(fout);
				end
			end
		else
			% write all data into a single file, separated by double-empty lines (gnuplot style)
			make_parent(dataFile);
			fout = fopen(dataFile, 'w');
			fprintf(fout, '# %s\n#\n', supertitle_sanitized);
			for j=1:N
				fprintf(fout,'# Data index %d\n# Depth profile: %s\n# time_depth\ttime\tdepth\t%s\n', j-1, titles{j}, cblabels{j});
				for ti=1:NT
					for zi=1:NZ
						fprintf(fout, '%g_%g\t %g\t%g\t%g\n', times(ti), depths(zi), times(ti), depths(zi), values(ti,zi,j)); 
					end
					fprintf(fout, '\n');
				end
				fprintf(fout, '\n\n');
			end
			fclose(fout);
		end
	end
end


% Ternary operator for C-like conditional expressions
function c = ternary(condition, a , b)
    if condition
        c = a;
    else
        c = b;
    end
end


% create a parent directory, if it does not exit
function make_parent(path)
	[parent,~,~] = fileparts(path);
	if(not(exist(parent, 'dir')))
		mkdir(parent);
	end
end


% create a Red-White-Blue color palette in MATLAB format
% coarse colors are taken from MATLAB's redbluecmap, and interpolated onto a finer grid
% Note that the Red-White-Blue colormap is not gray-scale safe
function palette = colormap_red_white_blue(Ncolors)
	if(nargin<1)
		Ncolors = size(get(gcf,'colormap'),1); 
	end
	coarse_palette = [	0.0156   0.1875   0.3789;
						0.1289   0.3984   0.6719;
						0.2617   0.5742   0.7617;
						0.5703   0.7695   0.8672;
						0.8164   0.8945   0.9375;
						0.9648   0.9648   0.9648;
						0.9883   0.8555   0.7773;
						0.9531   0.6445   0.5078;
						0.8359   0.3750   0.3008;
						0.6953   0.0938   0.16797;
						0.4023   0.0000   0.12109];
	Ncoarse   	= size(coarse_palette,1);
	coarse_x	= linspace(0,1,Ncoarse);
	coarse_dx 	= 1.0/(Ncoarse-1);
	palette		= zeros(Ncolors,3);
	for r=1:Ncolors
		x 	 = (r-1)/(Ncolors-1);
		left = 1+floor(x/coarse_dx);
		if(left==Ncoarse)
			palette(r,:) = coarse_palette(left,:);
		else
			right = min(left+1,Ncoarse);
			palette(r,:) = coarse_palette(left,:) + ((x-coarse_x(left))/(coarse_x(right)-coarse_x(left))) * (coarse_palette(right,:)-coarse_palette(left,:));
		end
	end
end




function s = sanitize_string_for_latex(s)
	s = strrep(s, '_', '.');
	s = strrep(s, '#', '\#');
end

