% LICENSE AGREEMENT
% - - - - - - - - -
% Copyright (c) 2022, Stilianos Louca
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
%
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in
%      the documentation and/or other materials provided with the distribution
%    * Neither the name of the original author (Stilianos Louca), nor the names
%      of any other contributors may be used to endorse or promote products derived
%      from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%
%
% DESCRIPTION
% - - - - - -
% This MATLAB code can be used to conveniently plot the results of an FRT simulation for a 1D system (e.g., a laterally homogeneous water column)
%
% - - - - - - - - -
% Tested on MATLAB R2019b - R2022a
% - - - - - - - - -
%
function plotFRTresults(times,...
						depths,...
						metabolites,...
						genes,...
						concentrations,...
						net_sources,...
						reaction_rates,...
						energy_rates,...
						deltaLogK,...
						ndeltaLogK,...
						tendencies,...
						results_dir,...
						verbose_prefix)
	PLOT_WIDTH			= 3;	% inches
	PLOT_HEIGHT			= 3;	% inches
	WIDE_PLOT_WIDTH		= 7;	% inches
	PLOT_CONTOUR_LEVELS	= 20;	% value resolution for spatiotemporal contour plots
	
	NT = length(times);
	NZ = length(depths);
	NR = length(genes);
	NM = length(metabolites);

	fprintf('%sPlotting FRT-predicted metabolite concentrations over spacetime..\n',verbose_prefix);
	plotSpatiotemporal(	times, ...
						depths, ...
						concentrations, ...
						{},...
						[1 1 1e6],...
						metabolites, ...
						metabolites, ...
						'day', ...
						'depth (m)',...
						'concentration ($\rm\mu M$)', ...
						false,... 		% logarithmicColorMap
						NaN,... 		% min_cbvalues
						strcat(results_dir,'/metabolite_concentrations.pdf'),...
						sprintf('FRT-predicted metabolite concentrations'), ...
						WIDE_PLOT_WIDTH, ...
						PLOT_HEIGHT, ...
						PLOT_CONTOUR_LEVELS, ...
						false, ... % equalizeColorRange
						strcat(results_dir,'/plot_data/metabolite_concentrations.tsv'),...
						true);

	fprintf('%sPlotting FRT-predicted final metabolite concentrations (depth profiles)..\n',verbose_prefix);
	plotProfiles(	reshape(concentrations(end,:,:),[NZ NM]),...
					depths, ...	
					{}, ... 	% data
					1e6, ... 	% rescales
					[], ...		% zrange
					metabolites, ...
					metabolites, ...
					'concentration ($\rm\mu M$)', ...
					strcat(results_dir,'/metabolite_concentrations_final.pdf'),...
					sprintf('FRT-predicted final metabolite concentrations (t=%g)',times(end)), ...
					0,... 		% minPlotValue
					NaN, ... 	% refValue
					false, ...	% logarithmic
					false,...	% show_points
					PLOT_WIDTH, ...
					PLOT_HEIGHT, ...
					false,... 		% showDepthIntegrated
					true,... 		% showExtrema
					false,...		% showR2P2
					NaN, ... 		% data_colors
					strcat(results_dir,'/plot_data/metabolite_concentrations_final.tsv'),...
					true); 			% split_dataFile

	fprintf('%sPlotting FRT-predicted net metabolite sources over spacetime..\n',verbose_prefix);
	plotSpatiotemporal(	times, ...
						depths, ...
						net_sources, ...
						{},...
						[1 1 1e6],...
						metabolites, ...
						metabolites, ...
						'day', ...
						'depth (m)',...
						'net source rate ($\rm\mu M/day$)', ...
						false,... 		% logarithmicColorMap
						NaN,... 		% min_cbvalues
						strcat(results_dir,'/net_metabolite_sources.pdf'),...
						sprintf('FRT-predicted metabolite concentrations'), ...
						WIDE_PLOT_WIDTH, ...
						PLOT_HEIGHT, ...
						PLOT_CONTOUR_LEVELS, ...
						false, ... % equalizeColorRange
						strcat(results_dir,'/plot_data/net_metabolite_sources.tsv'),...
						true);

	fprintf('%sPlotting FRT-predicted time-averaged net metabolite sources..\n',verbose_prefix);
	plotProfiles(	reshape(nanmean(net_sources, 1),[NZ NM]),...
					depths, ...
					{}, ... 				% data
					1e6, ...
					[depths(1),depths(end)], ...
					metabolites, ...
					metabolites, ...
					'rate ($\rm\mu M/day$)', ...
					strcat(results_dir,'/net_metabolite_sources_time_averaged.pdf'),...
					sprintf('FRT-predicted net metabolite source rates, averaged over time'), ...
					NaN,... 		% min x value
					NaN, ... 		
					false, ...		% logarithmic
					false, ...		% show_points
					PLOT_WIDTH, ...
					PLOT_HEIGHT, ...
					true, ... 	% showDepthIntegrated
					false, ...	% showExtrema
					false, ... 	% showR2P2
					NaN, ...	% data_colors
					strcat(results_dir,'/plot_data/net_metabolite_sources_time_averaged.tsv'),...
					true);

	fprintf('%sPlotting FRT-predicted final net metabolite sources..\n',verbose_prefix);
	plotProfiles(	reshape(net_sources(end,:,:),[NZ NM]),...
					depths, ...
					{}, ... 				% data
					1e6, ...
					[depths(1),depths(end)], ...
					metabolites, ...
					metabolites, ...
					'rate ($\rm\mu M/day$)', ...
					strcat(results_dir,'/net_metabolite_sources_final.pdf'),...
					sprintf('FRT-predicted final net metabolite source rates (t=%g)',times(end)), ...
					NaN,... 		% min x value
					NaN, ... 		
					false, ...		% logarithmic
					false, ...		% show_points
					PLOT_WIDTH, ...
					PLOT_HEIGHT, ...
					true, ... 	% showDepthIntegrated
					false, ...	% showExtrema
					false, ... 	% showR2P2
					NaN, ...	% data_colors
					strcat(results_dir,'/plot_data/net_metabolite_sources_final.tsv'),...
					true);

	fprintf('%sPlotting FRT-predicted reaction tendencies..\n',verbose_prefix);
	plotSpatiotemporal(	times, ...
						depths, ...
						tendencies, ...
						{},...
						[1 1 1],...
						genes, ...
						genes, ...
						'day', ...
						'depth (m)',...
						'tendency', ...
						false,... 	% logarithmicColorMap
						NaN,... 	% min_cbvalues
						strcat(results_dir,'/reaction_tendencies.pdf'),...
						sprintf('FRT-predicted reaction tendencies'), ...
						WIDE_PLOT_WIDTH, ...
						PLOT_HEIGHT, ...
						PLOT_CONTOUR_LEVELS, ...
						false, ... % equalizeColorRange
						strcat(results_dir,'/plot_data/reaction_tendencies.tsv'),...
						true);
						
	fprintf('%sPlotting FRT-predicted deltaLogK..\n',verbose_prefix);
	plotSpatiotemporal(	times, ...
						depths, ...
						deltaLogK, ...
						{},...
						[1 1 1],...
						genes, ...
						genes, ...
						'day', ...
						'depth (m)',...
						'deltaLogK', ...
						false,... 	% logarithmicColorMap
						NaN,... 	% min_cbvalues
						strcat(results_dir,'/reaction_deltaLogK.pdf'),...
						sprintf('FRT-predicted reaction deltaLogK'), ...
						WIDE_PLOT_WIDTH, ...
						PLOT_HEIGHT, ...
						PLOT_CONTOUR_LEVELS, ...
						false, ... % equalizeColorRange
						strcat(results_dir,'/plot_data/reaction_deltaLogK.tsv'),...
						true);

	fprintf('%sPlotting FRT-predicted normalized deltaLogK..\n',verbose_prefix);
	plotSpatiotemporal(	times, ...
						depths, ...
						ndeltaLogK, ...
						{},...
						[1 1 1],...
						genes, ...
						genes, ...
						'day', ...
						'depth (m)',...
						'normalized deltaLogK', ...
						false,... 	% logarithmicColorMap
						NaN,... 	% min_cbvalues
						strcat(results_dir,'/reaction_normalized_deltaLogK.pdf'),...
						sprintf('FRT-predicted reaction normalized deltaLogK'), ...
						WIDE_PLOT_WIDTH, ...
						PLOT_HEIGHT, ...
						PLOT_CONTOUR_LEVELS, ...
						false, ... % equalizeColorRange
						strcat(results_dir,'/plot_data/reaction_normalied_deltaLogK.tsv'),...
						true);
	
	fprintf('%sPlotting FRT-predicted reaction rates..\n',verbose_prefix);
	plotSpatiotemporal(	times, ...
						depths, ...
						reaction_rates, ...
						{},...
						[1 1 1e6],...
						genes, ...
						genes, ...
						'day', ...
						'depth (m)',...
						'rate ($\rm\mu M/day$)', ...
						false,... 	% logarithmicColorMap
						NaN,... 	% min_cbvalues
						strcat(results_dir,'/reaction_rates.pdf'),...
						sprintf('FRT-predicted reaction rates'), ...
						WIDE_PLOT_WIDTH, ...
						PLOT_HEIGHT, ...
						PLOT_CONTOUR_LEVELS, ...
						false, ... % equalizeColorRange
						strcat(results_dir,'/plot_data/reaction_rates.tsv'),...
						true);

	fprintf('%sPlotting FRT-predicted normalized reaction rates..\n',verbose_prefix);
	normalized_reaction_rates = bsxfun(@rdivide, reaction_rates, reshape(max(abs(reaction_rates),[],2),[size(reaction_rates,1),1,size(reaction_rates,3)]));
	plotSpatiotemporal(	times, ...
						depths, ...
						normalized_reaction_rates, ...
						{},...
						[1 1 1],...
						genes, ...
						genes, ...
						'day', ...
						'depth (m)',...
						'normalized rate (rate/max_abs)', ...
						false,... 	% logarithmicColorMap
						NaN,... 	% min_cbvalues
						strcat(results_dir,'/normalized_reaction_rates.pdf'),...
						sprintf('FRT-predicted normalized reaction rates'), ...
						WIDE_PLOT_WIDTH, ...
						PLOT_HEIGHT, ...
						PLOT_CONTOUR_LEVELS, ...
						false, ... % equalizeColorRange
						strcat(results_dir,'/plot_data/normalized_reaction_rates.tsv'),...
						true);

	fprintf('%sPlotting FRT-predicted depth-integrated reaction rates..\n',verbose_prefix);
	integrated_reaction_rates = reshape(sum(bsxfun(@times, reshape(diff(depths), [1,NZ-1,1]), 0.5 * (reaction_rates(:,2:end,:)+reaction_rates(:,1:(end-1),:))),2),[NT NR]);
	plotTimeSeries(	integrated_reaction_rates,...
					times, ...
					{}, ... 
					NaN, ... 	% time_rescaling
					NaN, ... 	% value_rescalings
					[], ...		% time_range
					[], ...		% value_range
					genes,...				% names
					genes, ...				% either a single string or a 1D cell array of strings and of size N
					'day', ...
					'depth-integrated rate ($\rm kmol\cdot m^{-2}\cdot d^{-1}$)', ...
					strcat(results_dir,'/reaction_rates_depth_integrated.pdf'),...
					'FRT-predicted depth-integrated reaction rates over time', ...
					0, ... 			% refValue
					false, ...		% logarithmic
					false, ...		% reverse_y
					WIDE_PLOT_WIDTH, ...
					PLOT_HEIGHT, ...
					true,...	 	% showMean
					false,...		% showR2
					NaN, ... 		% data_colors
					strcat(results_dir,'/plot_data/reaction_rates_depth_integrated.tsv'),...
					true);

	fprintf('%sPlotting FRT-predicted time-averaged reaction rates..\n',verbose_prefix);
	plotProfiles(	reshape(nanmean(reaction_rates, 1),[NZ NR]),...
					depths, ...
					{}, ... 				% data
					1e6, ...
					[depths(1),depths(end)], ...
					genes, ...
					genes, ...
					'rate ($\rm\mu M/day$)', ...
					strcat(results_dir,'/reaction_rates_time_averaged.pdf'),...
					sprintf('FRT-predicted reaction rates, averaged over time'), ...
					NaN,... 		% min x value
					NaN, ... 		
					false, ...		% logarithmic
					false, ...		% show_points
					PLOT_WIDTH, ...
					PLOT_HEIGHT, ...
					true, ... 	% showDepthIntegrated
					false, ...	% showExtrema
					false, ... 	% showR2P2
					NaN, ...	% data_colors
					strcat(results_dir,'/plot_data/reaction_rates_time_averaged.tsv'),...
					true);

	fprintf('%sPlotting FRT-predicted final reaction rates..\n',verbose_prefix);
	plotProfiles(	reshape(reaction_rates(end,:,:),[NZ NR]),...
					depths, ...
					{}, ... 				% data
					1e6, ...
					[depths(1),depths(end)], ...
					genes, ...
					genes, ...
					'rate ($\rm\mu M/day$)', ...
					strcat(results_dir,'/reaction_rates_final.pdf'),...
					sprintf('FRT-predicted final reaction rates (t=%g)',times(end)), ...
					NaN,... 		% min x value
					NaN, ... 		
					false, ...		% logarithmic
					false, ...		% show_points
					PLOT_WIDTH, ...
					PLOT_HEIGHT, ...
					true, ... 	% showDepthIntegrated
					false, ...	% showExtrema
					false, ... 	% showR2P2
					NaN, ...	% data_colors
					strcat(results_dir,'/plot_data/reaction_rates_final.tsv'),...
					true);

	fprintf('%sCompressing FRT plot data files..\n',verbose_prefix);
	compress_dir_contents(strcat(results_dir,'/plot_data'));
end



% recursively gzip all files in a given folder, removing the original files
function compress_dir_contents(dirpath)
	if(exist(dirpath, 'dir'))
		compressed_files = gzip(dirpath);
		for f=1:length(compressed_files)
			compressed_file = compressed_files{f};
			uncompressed_file = compressed_file(1:end-3);
			if(exist(uncompressed_file, 'file'))
				delete(uncompressed_file);
			end
		end
	end
end
