% LICENSE AGREEMENT
% - - - - - - - - -
% Copyright (c) 2019, Stilianos Louca
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
%
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in
%      the documentation and/or other materials provided with the distribution
%    * Neither the name of the original author (Stilianos Louca), nor the names
%      of any other contributors may be used to endorse or promote products derived
%      from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%
% - - - - - - - - -
% Tested on MATLAB R2015b.
% - - - - - - - - -
%
% DESCRIPTION
% - - - - - -
% Plot time series for N observables
% profiles(:,j) is time series of j-th observable across times[].
% titles{} is either a cell array of N strings or a single string. Similarly for xlabels{} and ylabels{}.
% data{} is either empty or a cell array of size N. Each entry can be a table (L*2, one table per observable) or NaN (no data avail).
function plotTimeSeries(profiles,...
						times, ...				% 1D vector of size NT, listing times in ascending order
						data, ... 
						time_rescaling, ... 	% either [] or NaN or a single number, specifying an optional rescaling factor for the time axis in the data and in the profiles.
						value_rescalings, ... 	% either [] or NaN or a single number or a vector of size N, giving rescaling factors both for data and profiles (value/Y axis).
						time_range, ...			% either an empty vector, or [min,max]. This range applies to the potentially rescaled times.
						value_range, ...		% either an empty vector, or [min,max]. Either min and/or max may also be NaN. This range applies to the potentially rescaled values.
						names,...				% 1D cell array of length N, specifying brief names for the various plots. Used e.g. to generate file names. For example {'NH4', 'O2', 'SO4'}.
						titles, ...				% either a single string or a 1D cell array of strings and of size N
						xlabels, ...			% either a single string or a 1D cell array of strings and of size N
						ylabels, ...			% either a single string or a 1D cell array of strings and of size N
						plotFile,...
						supertitle, ...
						refValue, ... 			% can be NULL or a real number or a vector of numbers and NaNs. Defines reference value to add to plots (as horizontal line).
						logarithmic, ...
						reverse_y, ...
						plotWidth, ...
						plotHeight, ...
						showMean,...	 	% if true, the time-integrated mean for each plot will be shown as a comment in the title
						showR2,...			% if true, and if applicable, the coefficient of determination (R2) to the data will be shown as a comment in the title
						data_colors, ... 		% can be NaN or a single string or a cell array of colors for each data set
						dataFile,... 			% either a path to a single text file, or a basepath to multiple text files (depending on split_dataFile), for saving data. Can be '' (empty).
						split_dataFile)			% if true, then the data will be written in separate files, one file per plot per data type
	N  = size(profiles,2);
	NT = size(profiles,1);
	
	% pre-allocate some arrays
	meanValues = repmat(NaN,N,1);
	stdValues  = repmat(NaN,N,1);

	% re-format titles & labels if needed
	if(~iscell(names))
		names = repmat({names},N,1);
	end
	if(~iscell(titles))
		titles = repmat({titles},N,1);
	end
	if(~iscell(xlabels))
		xlabels = repmat({xlabels},N,1);
	end
	if(~iscell(ylabels))
		ylabels = repmat({ylabels},N,1);
	end
	if(length(value_rescalings)==1)
		value_rescalings = repmat(value_rescalings,[N 1]);
	end
	if(not(isempty(dataFile)) & split_dataFile && (endsWith(dataFile,'.tsv') || endsWith(dataFile,'.txt')))
		% user provided a TSV file path even though they should have provided a basepath, so convert it into a basepath
		dataFile = dataFile(1:end-3);
	end

	% apply optional time rescaling (applies to all plots)
	if((length(time_rescaling)==1) && not(isnan(time_rescaling)))
		times = times * time_rescaling;
		for j=1:length(data)
			if(size(data{j},2)>1)
				data{j}(:,1) = time_rescaling * data{j}(:,1);
			end
		end
	end
		
	% geometry for sub-plot distribution
	[nrows ncols] = num2RowsColumns(N, 0.6,0.2,3);
	titleAreaV	= 0.05 * (1+length(strfind(supertitle,sprintf('\n')))); % reserve title space
	plotAreaV 	= 1-titleAreaV;
	
	% initialize figure window (make invisible since we plot everything to pdf)
	fig = figure('Visible','off','PaperPosition',[0 0 ncols*plotWidth nrows*plotHeight],'PaperSize',[ncols*plotWidth nrows*plotHeight]);
	
	if(size(time_range,1)>0) 
		time_epsilon = (time_range(2) - time_range(1))*1e-3;
	end
	
	% iterate through sub plots
	for row=1:nrows
		for col=1:ncols
			j = (row-1)*ncols+col;
			if(j>N) break; end
			
			hasData = ((length(data)>=j) && (size(data{j},2)>1) && (size(data{j},1)>0));
			
			% rescale values in profiles & data if necessary
			if((length(value_rescalings)>0) && not(isnan(value_rescalings(j))))
				profiles(:,j) = profiles(:,j) * value_rescalings(j);
				if(hasData) data{j}(:,2) = data{j}(:,2) * value_rescalings(j); end
			end

			% mannually place subplot
			% Note: If subplot overlaps previous subplots, the older ones will be deleted, unless overlap is exact! (this is an idiotic matlab 'feature')
			subplot_position = [(0.15+col-1)/ncols plotAreaV*(0.25+nrows-row)/nrows 0.7/ncols plotAreaV*0.6/nrows];
			subplot('Position',subplot_position);
			
			% plot profile
			if(all(isnan(profiles(:,j))))
				% all values are NaN, so just show an error message as plot title
				title(sprintf('%s: All NaN',titles{j}),'fontsize',9);
				continue;
			end
			plot(times,profiles(:,j),'LineWidth',1);
			
			% set y-axis limits
			if((length(value_range)>0) & not(isnan(value_range(1))) & not(isnan(value_range)))
				ylim(value_range);
			else
				minValue = min(profiles(:,j));
				maxValue = max(profiles(:,j));
				if(hasData)
					if(size(time_range,1)>0) shownData = find((data{j}(:,1)<=time_range(end)+time_epsilon) & (data{j}(:,1)>=time_range(1)-time_epsilon)); 
					else shownData = 1:size(data{j},1); end
					if(length(shownData)>0)
						minValue = min(minValue, min(data{j}(shownData,2))); 
						maxValue = max(maxValue, max(data{j}(shownData,2))); 
					end
				end
				L = (maxValue-minValue);
				minValue = ternary(minValue<0, minValue-0.1*L, max(0,minValue-0.1*L));
				maxValue = ternary(maxValue>0, maxValue+0.1*L, min(0,maxValue+0.1*L));
				if(minValue==maxValue)
					if(minValue==0)
						maxValue = 1;
					else
						minValue = minValue-0.01*abs(minValue);
						maxValue = maxValue+0.01*abs(maxValue);
					end
				end
				if((length(value_range)>0) & not(isnan(value_range(1))))
					minValue = value_range(1);
				end
				if((length(value_range)>0) & not(isnan(value_range(2))))
					maxValue = value_range(2);
				end
				ylim([minValue maxValue]);
			end
			if(logarithmic) set(gca, 'YScale', 'log'); end
			if(reverse_y) set(gca,'YDir','reverse'); end
			
			% set time-axis limits
			if(size(time_range,1)>0) 
				if(hasData)
					xlim([time_range(1)-time_epsilon,time_range(2)+time_epsilon]); % need to 'widen' up plot area a bit to make sure all relevant data points are shown
				else
					xlim(time_range);
				end
			end
			
			% set title
			thisTitle = titles{j};
			if(showMean)
				meanValues(j) = integrateVector(reshape(times,[NT 1]), reshape(profiles(:,j),[NT 1]), true);
				stdValues(j)  = sqrt(integrateVector(reshape(times,[NT 1]), reshape((profiles(:,j)-meanValues(j)).^2,[NT 1]), true));
				thisTitle 	  = sprintf('%s\ntemporal average = %.4g +/- %.2g s.d.', thisTitle, meanValues(j), stdValues(j));
			end
			if(showR2 & hasData)
				R2 = getProfileR2(times, profiles(:,j), data{j}(:,1), data{j}(:,2));
				thisTitle = sprintf('%s\nR2=%.3g', thisTitle, R2);
			end
			title(sanitize_string_for_latex(thisTitle));
			
			xlabel(sanitize_string_for_latex(xlabels{j}),'Interpreter','latex');
			ylabel(sanitize_string_for_latex(ylabels{j}),'Interpreter','latex');
			set(gca,'fontsize',8);
			
			% overlay reference value
			if(length(refValue)>1 | not(isnan(refValue)))
				if(length(refValue)==1) rv = refValue;
				elseif(length(refValue)>=j) rv = refValue(j);
				else rv = NaN;
				end
				if(not(isnan(rv)))
					hold on
					plot(xlim,[rv rv],'Color',[0.5 0.5 0.5]);
					hold off
				end
			end
			
			% overlay data
			if(hasData)
				if(iscell(data_colors)) color = data_colors{j};
				elseif(isnan(data_colors)) color = 'red';
				else color = data_colors; end
				hold on
				scatter(data{j}(:,1), data{j}(:,2), 6, color);
				hold off
			end
		end
	end
		
	%create super title
	%this should come after subplot(), because the latter has a bad tendency to delete previous figure elements
	annotation('textbox', [0 (1-titleAreaV) 1 titleAreaV], 'String', sanitize_string_for_latex(supertitle), 'EdgeColor', 'none', 'HorizontalAlignment', 'center', 'VerticalAlignment', 'top','Interpreter','latex');

	make_parent(plotFile);
	print(fig,'-dpdf',plotFile);
	
	% Write data to file
	if(not(isempty(dataFile)))
		supertitle_sanitized = strrep(supertitle, sprintf('\n'), sprintf('\n#   '));
		if(split_dataFile)
			% write each dataset into a separate file (one file per plot per data type)
			for j=1:N
				% write j-th profile
				filepath = strcat(dataFile,names{j},'.profile.tsv');
				make_parent(filepath);
				fout = fopen(filepath, 'w');
				fprintf(fout, '# %s\n# Time series profile for: %s, %s\n# time\t%s\n', supertitle_sanitized, titles{j}, ylabels{j}, names{j});
				for ti=1:NT
					fprintf(fout, '%g\t%g\n', times(ti), profiles(ti,j)); 
				end
				fclose(fout);
				% write overlay data (if available)
				if(length(data)>=j)
					if(size(data{j},2)>1)
						fout = fopen(strcat(dataFile,names{j},'.scatter.tsv'), 'w');
						fprintf(fout, '# %s\n# Scatterdata for: %s, %s\n# time\t%s\n', supertitle_sanitized, titles{j}, ylabels{j}, names{j});
						for di=1:size(data{j},1)
							fprintf(fout, '%g\t%g\n', data{j}(di,1), data{j}(di,2));
						end
						fclose(fout);
					end
				end
			end
		else
			% write all data into a single file, separated by double-empty lines (gnuplot style)
			index = 0;
			make_parent(dataFile);
			fout = fopen(dataFile, 'w');
			fprintf(fout, '# %s\n#\n', supertitle_sanitized);
			for j=1:N
				% write profile
				fprintf(fout,'# Data index %d\n# Time series: %s\n', index, titles{j});
				if(showMean) fprintf(fout,'# Mean value: %.10g +/- %.2g s.d.\n',meanValues(j),stdValues(j)); end
				fprintf(fout,'# Data columns: %s %s\n', xlabels{j}, ylabels{j});
				for ti=1:NT
					fprintf(fout, '%g\t%g\n', times(ti), profiles(ti,j)); 
				end
				fprintf(fout, '\n\n'); index = index + 1;
				% write overlay data (if available)
				if(length(data)>=j)
					if(size(data{j},2)>1)
						fprintf(fout, '# Data index %d\n# Data scatter points: %s\n# Data columns: %s\t%s\n', index, titles{j}, xlabels{j}, ylabels{j});
						for di=1:size(data{j},1)
							fprintf(fout, '%g\t%g\n', data{j}(di,1), data{j}(di,2));
						end
						fprintf(fout, '\n\n'); index = index + 1;
					end
				end
			end
		
			% include table with mean values, if needed
			if(showMean)
				fprintf(fout,'# Data index: %d\n# Mean values of all time series\n# series\tmean_value\tstd_value\n',index);
				for j=1:N
					fprintf(fout,'%s\t%.10g\t%.10g\n',titles{j},meanValues(j),stdValues(j));
				end
				fprintf(fout, '\n\n'); index = index + 1;
			end
			fclose(fout);
		end
	end
end



% Ternary operator for C-like conditional expressions
function c = ternary(condition, a , b)
    if condition
        c = a;
    else
        c = b;
    end
end



% integrate a scalar y along x
% x is a vector of length L
% y is a vector of length L
% average is a boolean
% ignores NaNs
function I = integrateVector(x, y, average)
	valids = find((~isnan(x)) & (~isnan(y)));
	if(length(valids)<=1)
		I = NaN;
		return;
	end
	x 		= x(valids);
	y 		= y(valids);
	dx 		= diff(x);
	ym 		= (y(2:end)+y(1:end-1))/2;
	I 		= squeeze(sum(bsxfun(@times,ym,reshape(dx,length(dx),1)), 1));
	if(average)
		I = I/(x(end)-x(1));
	end
end



function R2 = getProfileR2(model_z, model_values, data_z, data_values)
	% interpolate model values at data z
	model_values_at_data_z = interp1q(model_z, model_values, data_z);
	valids = not(isnan(model_values_at_data_z)); 	% some depths might return NaN in interpolation, e.g. when outside of interpolation domain. 
	if(not(any(valids)))
		% no valid model or data values available at any data depth
		R2 = NaN; 
		return; 
	end
	count	= sum(valids);
	
	SSR		= sum((data_values(valids) - model_values_at_data_z(valids)).^2);
	SST		= sum(data_values(valids).^2) - (sum(data_values(valids)))^2/count;
	R2 		= 1 - SSR/SST;
end


% create a parent directory, if it does not exit
function make_parent(path)
	[parent,~,~] = fileparts(path);
	if(not(exist(parent, 'dir')))
		mkdir(parent);
	end
end



function s = sanitize_string_for_latex(s)
	s = strrep(s, '_', '.');
	s = strrep(s, '#', '\#');
end

