% LICENSE AGREEMENT
% - - - - - - - - -
% Copyright (c) 2022, Stilianos Louca
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
%
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in
%      the documentation and/or other materials provided with the distribution
%    * Neither the name of the original author (Stilianos Louca), nor the names
%      of any other contributors may be used to endorse or promote products derived
%      from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%
%
% DESCRIPTION
% - - - - - -
% This MATLAB code simulates a spatiotemporal Fast-Reaction-Transport (FRT) model for 1-dimensional systems.
% All reactions are assumed to occur infinitely fast (more precisely, much faster than physical transport or environmental change).
% The function can account for a lateral basin area that varies with depth.
%
% The PDE solved has the form:
%		p_tC = p_z(D*p_zC - v*C) + (1/A)*(p_zA) * (D*p_zC - v*C) + stoichiometry * fast_reaction_rates
%
% Main inputs:
%	The spatiotemporal grid for which to return solutions
%	Boundary conditions for each metabolite at the top and bottom (Dirichlet/fixed-concentration or Robin/fixed-flux conditions)
%   The diffusivity and advection speed for each metabolite
%	Stoichiometric matrix of all reactions
%	Standard Gibbs free energy for each reaction
% Main outputs:
%	The predicted metabolite concentrations on the requested spatiotemporal grid
%
% Canonical units assumed for each variable:
%	Depth (location): meters
%	Time: days
%	Diffusivity: m^2/day
%	Advection speed: m/day
%	Relative dilution slope (relative slope of change of lateral basin area over depth): unitless, (1/A) * dA/dz
%	Metabolite concentrations: mol/L
%	Reaction rate densities: mol/L/day
%	Temperature: degrees Celsius
%	Reaction equilibrium constants
%
% FRT models are related to ERT models in that they assume reactions to be infinitely fast. However, FRT does not assume that all reactions are at thermodynamic equilibrium.
% For a background on ERT models in geochemistry see:
%	Rubin (1983). Transport of reacting solutes in porous media: Relation between mathematical nature of problem formulation and chemical nature of reactions. Water Resources Research. 19:1231-1252.
%	Lichtner (1985). Continuum model for simultaneous chemical reactions and mass transport in hydrothermal systems. Geochimica et Cosmochimica Acta. 49:779-800.
%	Neretnieks (1993). Diffusion, Flow and Fast Reactions in Porous Media. in "Migration and fate of pollutants in soils and subsoils". Editors Petruzzelli and Helfferich. Springer. pages 329-354.
%	
%
% - - - - - - - - -
% Code name and version: FRT 2022
% Tested on MATLAB R2019b - R2022a
% - - - - - - - - -
%
function [	concentrations,...	% NT x NZ x NM array listing the predicted metabolite concentrations
			net_sources,...		% NT x NZ x NM array listing the net source rates for each metabolite
			reaction_rates,...	% NT x NZ x NR array listing reaction rates. Note that these may not be fully identifiable, since multiple alternative reaction rates can yield the same metabolite fluxes.
			energy_rates,...	% NT x NZ array listing Gibbs free energy rates at each spacetime point and summed over all reactions (-\sum_r DeltaG_r(t,z)*rate_r(t,z)). A positive value corresponds to exergonic reactions. These energy rates account for the excluded metabolites, based on the provided restricted equilibrium constants.
			deltaLogK,...		% NT x NZ x NR numeric array listing the deviations from equilibrium
			ndeltaLogK,...		% NT x NZ x NR numeric array listing deltaLogK/abs(logK)
			tendencies,...		% NT x NZ x NR integer array listing the tendencies (directions) of the reactions (-1:reverse, 1:forward, 0:equilibrium, NaN:undefined)
			errormsg] = runFRTmodel(gridT,...					% 1D numeric vector of size NT, specifying the time points for which to calculate SERT solutions. A higher resolution just means a higher resolution in the returned solution grids, but does not substantially impact accuracy nor computing time.
									gridZ,...					% 1D numeric vector of size NZ, specifying the depths for which to calculate SERT solutions, including the boundary depths zA and zB. A higher resolution means more accurate calculations, but roughly proportionally more computing time.
									NM, ...						% integer, number of metabolites
									NR, ...						% integer, number of reactions
									metabolite_names,...		% cell array of size NM, specifying metabolite names. Mostly used for progress reports and error messages
									reaction_names,...			% cell array of size NR, specifying reaction names. Mostly used for progress reports and error messages
									BCtypesUp, ...				% 1D integer array of size NM, listing the type of boundary conditions for each metabolite, at the top boundary (0:Dirichlet/fixed value, 1:Robin/fixed influx)
									BCtypesDown, ...			% 1D integer array of size NM, listing the type of boundary conditions for each metabolite, at the bottom boundary (0:Dirichlet/fixed value, 1:Robin/fixed influx)
									BCvaluesUp, ...				% cell-array of size NM, listing time-dependent functionals, specifying boundary concentrations (if BC type==0) or influxes (if BC type==1) of metabolites at zA (top boundary) at any time t
									BCvaluesDown, ...			% cell-array of size NM, listing time-dependent functionals, specifying boundary concentrations (if BC type==0) or influxes (if BC type==1) of metabolites at zB (bottom boundary) at any time t
									ICs,...						% cell-array of size NM, listing depth-dependent functionals, specifying initial concentrations of metabolites at gridT(1) (start time)
									diffusivities, ... 			% 1D cell array of size 1 or NM, defining the vertical diffusivities of metabolites (scalar-valued) at any (t,z). If of size 1, then all metabolites are assumed to have the same diffusivity. If of size 0 (empty cell array), then diffusivity is assumed to be zero for all metabolites.
									diffusivity_gradients,... 	% 1D cell array of size 1 or NM, defining the depth-derivative of the diffusivity (dD/dz) for each metabolite at any (t,z). If of size 1, then all metabolites are assumed to have the same diffusivity gradient. Only relevant if use_ficks_law==false. If use_ficks_law==false, you can pass an empty cell array ({}).
									advections, ... 			% 1D cell array of size 1 or NM, defining the advection velocities of metabolites at any (t,z). If of size 1, then all metabolites are assumed to have the same advection velocity. If of size 0 (empty cell array), then advection is assumed to be zero for all metabolites. Can also contain scalars instead of function handles, indicating a spatially-constant advection.
									dilution_slope,...			% depth-dependent function handle, specifying the relative slope of the lateral (cross-sectional) area as a function of z. This is (1/A)*dA/dz, where A is the lateral area (e.g. of a modeled ocean basin). Used to correct for geometric dilution effects. If you don't know what this is, you can have this function simply return 0.
									temperature,...				% scalar function handle, returning the temperature (in Celsius) for any (t,z). Should be vectorized in its arguments. Only needed for calculating energy fluxes, but not reaction rates or metabolite concentrations.
									use_ficks_law,...			% (boolean) whether to use Fick's law to model diffusion. In typical geobiological models this will be true.
									stoichiometry, ...			% 2D array of size NM x NR, specifying the stoichiometric matrix of reactions.
									log_equilibrium_constants,... % cell array of size NR, with each r-th entry being a vectorized function handle specifying the natural logarithm of the "restricted" equilibrium constant of the r-th reaction at each (t,z), i.e. restricted to the metabolites considered in the model. If some substrates/products involved in the reactions in reality (e.g., H2O, H+ etc) are omitted from the model, these restricted equilibrium constants must account for the fixed concentrations of those compounts, so that equilibrium indeed corresponds to the a true thermodynamic equilibrium (i.e., as if no metabolites had been omitted). Note that in general equilibrium constants depend on temperature (according to the Van't Hoff equation) and on pressure (although this dependence is usually weak). Alternatively, log_equilibrium_constants may be a single functional returning the logK for all NR reactions as a NR-sized vector; in that case, the function does not need to be vectorized in (t,z).
									reversible,....				% 1D boolean vector of size NR, specifying whether a reaction is reversible or not.
									check_nans_in_input,...		% (boolean) check if provided function handles (e.g. diffusivity or temperature) return NaN within the simulation domain. If check_nans_in_input==false, function handles are blindly assumed to not contain NaNs
									interleave_dt,...			% either a single positive scalar or a function handle (mapping time to dt, must be vectorized), specifying the time interval between successive reaction stages. If a function handle, it should specify the interleave_dt for any requested time point (this may be useful e.g. if interleave_dt needs to be shorter during certain time periods). A shorter interleave_dt will increase accuracy but also computing time, since the costly reaction-ODEs will need to be integrated more frequently.
									transport_dt,...			% either a single positive scalar or a function handle (mapping time to dt, must be vectorized), specifying the suggested initial time step to be used by the PDE solver for the transport part. If a function handle, it should specify the interleave_dt for any requested time point (this may be useful e.g. if transport_dt needs to be shorter during certain time periods).
									transport_relative_error,... 	% scalar or numeric array of size NM, relative error tolerance for PDE solver for the transport part
									reaction_relative_errors,...	% scalar or numeric array of size NM, listing the allowed relative error for all state variables of the reaction ODEs. Typical values are 0.01-0.0001.
									concentration_errors,...	% scalar or 1D numeric array of size NM, acceptable absolute error for metabolite concentrations when solving the transport PDE and the reaction ODEs. If some entries are NaN, they will be automatically determined based on the provided transport_relative_error & reaction_relative_errors and the typical concentrations encountered in the system.
									DeltaG_threshold,...		% either a positive scalar or numeric array of size NR, specifying the maximum acceptable DeltaG (kJ/mol) deviation from equilibrium for a reaction to proceed. For irreversible (i.e., biological) reactions this should encode the viability threshold (energy quantum, >=0, typically 10-50 kJ/mol), whereas for reversible (i.e., abiotic) reactions this should encode the numerical error tolerance (>0).
									kinetics,...				% optional function handle, taking as argument a vector of NM metabolite concentrations, and returning NR reaction rates. Only relative rates matter. If NaN, then reaction rates are assumed to be proportional to deltaLogK.
									reactions_max_runtime,...	% maximum allowed runtime (wall time in seconds) per fast reaction stage per grid cell. Fast reaction stages taking longer than this threshold will be stopped prematurely (returning whatever state they ended up in by that time). If NaN, this stopping criterion is ignored. Typical values for systems with 1-5 reactions are 1-10 seconds.
									parallel,...				% (boolean) whether to use multiple threads for parallel computation
									diagnostics,...				% (boolean) whether to print technical diagnostic messages along the way. For debugging purposes only.
									verbose, ...				% (boolean) if true, progress messages will be displayed
									verbose_prefix)				% (string) prefix for progress messages. E.g. "  "
	NT	= length(gridT);
	NZ	= length(gridZ);
	zU	= gridZ(1);
	zD	= gridZ(end);
	Rg	= 8.3144621e-3; % gas constant (kJ/mol/K)
	%deltaLogK_relative_epsilon = 0.001;

	% re-shape some input arrays to ensure proper format
	gridZ = reshape(gridZ, [NZ 1]);
	gridT = reshape(gridT, [NT 1]);
	[meshgridZ, meshgridT] = meshgrid(gridZ, gridT);
	if(length(concentration_errors)==1)
		concentration_errors = repmat(concentration_errors, [NM, 1]);
	end
	reversible = reshape(reversible, [NR,1]);

	% default return values
	concentrations	= NaN;
	net_sources		= NaN;
	reaction_rates	= NaN;
	energy_rates	= NaN;
	deltaLogK		= NaN;
	ndeltaLogK		= NaN;
	tendencies		= NaN;
	errormsg		= '';

	format long; % increase printing precision

	% basic error checking
	if(NT<2 | NZ<2)
		errormsg = 'Grid size must be at least 2x2'; return;
	end
	if(length(diffusivities)==1)
		if(isa(diffusivities,'function_handle')) diffusivities = {diffusivities}; end
		diffusivities = repmat(diffusivities, NM, 1);
	elseif(length(diffusivities) ~= NM)
		errormsg = sprintf('Incorrect number of diffusivities provided (%d); expected either 1 or %d',length(diffusivities),NM);
		return;
	end
	if(length(diffusivity_gradients)==0)
		if(not(use_ficks_law)) errormsg = 'Missing diffusivity gradient(s), needed since not using Ficks law'; return; end
		diffusivity_gradients = repmat({@(t,z) zeros(size(t))}, NM, 1);
	elseif(length(diffusivity_gradients)==1)
		if(isa(diffusivity_gradients,'function_handle')) diffusivity_gradients = {diffusivity_gradients}; end
		diffusivity_gradients = repmat(diffusivity_gradients, NM, 1);
	end
	if(length(advections)==1)
		if(isa(advections,'function_handle')) advections = {advections}; end
		advections = repmat(advections, NM, 1);
	elseif(length(advections) ~= NM)
		errormsg = sprintf('Incorrect number of advections provided (%d); expected either 1 or %d',length(advections),NM);
		return;
	end
	DeltaG_threshold = abs(DeltaG_threshold);
	if(length(DeltaG_threshold)==1)
		DeltaG_threshold = repmat(DeltaG_threshold, 1, NR);
	elseif(length(DeltaG_threshold)~=NR)
		errormsg = sprintf('Incorrect number of DeltaG_threshold provided (%d); expected either 1 or %d',length(DeltaG_threshold),NR);
		return;
	else
		DeltaG_threshold = reshape(DeltaG_threshold, [1, NR]);
		if(any(DeltaG_threshold<=0) & reversible)
			errormsg = sprintf('For reversible reactions, DeltaG_threshold must be strictly positive, for numerical reasons.');
			return;
		end
	end
	if(isa(log_equilibrium_constants,'function_handle'))
		logK = log_equilibrium_constants(gridT(1),gridZ(1));
		if(length(logK)~=NR)  errormsg = sprintf('logK returns array of wrong length (%d); expected length %d',length(logK),NR); return; end
		if((size(logK,1)~=1) & (size(logK,2)~=1))  errormsg = sprintf('logK returns multidimensional array, but expected 1D array'); return; end
	elseif(length(log_equilibrium_constants)~=NR) 
		errormsg = sprintf('logK has length %d, but expected length %d',length(log_equilibrium_constants),NR);
	end
	if(length(BCtypesUp)~=NM) errormsg = sprintf('Incorrect number of BCtypesUp provided (%d); expected %d',length(BCtypesUp),NM); return; end
	if(length(BCtypesDown)~=NM) errormsg = sprintf('Incorrect number of BCtypesDown provided (%d); expected %d',length(BCtypesDown),NM); return; end
	if(length(BCvaluesUp)~=NM) errormsg = sprintf('Incorrect number of BCvaluesUp provided (%d); expected %d',length(BCvaluesUp),NM); return; end
	if(length(BCvaluesDown)~=NM) errormsg = sprintf('Incorrect number of BCvaluesDown provided (%d); expected %d',length(BCvaluesDown),NM); return; end
	if(check_nans_in_input)
		% check if any of the input functionals return NaN on the spacetime grid
		if(verbose) fprintf('%sChecking validity of input functionals..\n', verbose_prefix); end
		if(any(isnan(dilution_slope(meshgridZ)))) errormsg = sprintf('Dilution slope is NaN at some depths'); return; end
		for m=1:NM
			if(any(isnan(diffusivities{m}(meshgridT,meshgridZ)))) errormsg = sprintf('Diffusivity is NaN in some spacetime points, for metabolite %s',metabolite_names{m}); return; end
			if(not(use_ficks_law) & any(isnan(diffusivity_gradients{m}(meshgridT,meshgridZ)))) errormsg = sprintf('Diffusivity gradient is NaN in some spacetime points, for metabolite %s',metabolite_names{m}); return; end
			if(any(isnan(advections{m}(meshgridT,meshgridZ)))) errormsg = sprintf('Advection speed is NaN in some spacetime points, for metabolite %s',metabolite_names{m}); return; end
		end
	end
	if(isa(interleave_dt,'function_handle'))
		dts = interleave_dt(gridT);
		if(any(isnan(dts))) errormsg = sprintf('Some interleave_dt values are NaN'); return; end
		if(any(dts<=0)) errormsg = sprintf('Some interleave_dt values are 0 or negative'); return; end
	end
	if(isa(transport_dt,'function_handle'))
		dts = transport_dt(gridT);
		if(any(isnan(dts))) errormsg = sprintf('Some transport_dt values are NaN'); return; end
		if(any(dts<=0)) errormsg = sprintf('Some transport_dt values are 0 or negative'); return; end
	end
			
	% define initial concentrations
	if(verbose) fprintf('%sSetting initial concentrations..\n', verbose_prefix); end
	concentrations = zeros(NT,NZ,NM);
	for m=1:NM
		concentrations(1,:,m) = ICs{m}(gridZ);
	end
	
	if(isa(kinetics,'function_handle'))
		if(any(reversible)) errormsg = sprintf('Kinetics are only allowed if all reactions are non-reversible. To include reversible reactions, explicitly specify both directions as separate reactions, each with their own kinetics.'); return; end
		test_rates = kinetics(concentrations(1,NZ/2,:));
		if(length(test_rates)~=NR) errormsg = sprintf('Kinetics returns incorrect number of reaction rates'); return; end
		if(any(isnan(test_rates))) errormsg = sprintf('Kinetics returns NaN reaction rates'); return; end
	end
			
	if(NT==1)
		return; % nothing more to be done
	end
	
	
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% solve fast-reaction-transport equations
	
	% prepare central-difference gradient operator (over depth) in matrix form
	gradient_operator = zeros(NZ,NZ);
	basis_vector = zeros(NZ,1);
	for c=1:NZ
		if(c>1) basis_vector(c-1) = 0; end
		basis_vector(c) = 1;
		gradient_operator(:,c) = gradient(basis_vector,gridZ);
	end
	gradient_operator = sparse(gradient_operator); % make sparse for more efficient matrix multiplication
		
	if(verbose) fprintf('%sSimulating FRT..\n', verbose_prefix); end
	deltaLogK 		= zeros([NT,NZ,NR]);
	ndeltaLogK 		= zeros([NT,NZ,NR]);
	tendencies 		= zeros([NT,NZ,NR]);
	net_sources		= zeros([NT,NZ,NM]);
	reaction_rates	= zeros([NT,NZ,NR]);
	energy_rates	= zeros([NT,NZ]);
	dAdz 			= dilution_slope(gridZ);
	D 				= zeros([NZ,NM]); % scratch storage for diffusivities
	V 				= zeros([NZ,NM]); % scratch storage for advection speeds
	logK 			= zeros([NZ,NR]); % scratch storage for reaction logK
	Ccurrent 		= reshape(concentrations(1,:,:),[NZ,NM]);
	Ctransported	= zeros([NZ,NM]); % scratch storage for intermediate concentrations vector, after the transport step
	Cnext			= zeros([NZ,NM]); % scratch storage for next concentrations vector
	Cparfor			= zeros([NZ-2,NM]); % scratch storage for shuffled concentrations from parfor loop
	Rparfor			= zeros([NZ-2,NR]); % scratch storage for shuffled reaction fluxes from parfor loop
	reaction_fluxes	= zeros([NZ,NR]); % scratch storage for reaction fluxes
	time 			= gridT(1);
	next_tg 		= 2; % next requested time-point for which the concentrations are to be computed
	warning('off','MATLAB:mir_warning_maybe_uninitialized_temporary'); % suppress this annoying warning about temporary variables inside parfor loops
	parfor_OK 		= true; % status of simulations inside parfor loops
	depth_shuffle 	= 1+randperm(NZ-2); % mapping for shuffling depths 2:(NZ-1), so that parfor randomly assigns depths to workers (as opposed to each worker getting a block of contiguous depths, which can correlate in their runtimes). This helps balance the load more evenly across parfor workers.
	interior		= 2:(NZ-1);
	while(time<gridT(end))
		% determine the currect interleave_dt (time until next reaction stage) and transport_dt (suggested time step for the transport stage)
		if(isa(interleave_dt,'function_handle'))
			idt = interleave_dt(time);
		else
			idt = interleave_dt;
		end
		if(isa(transport_dt,'function_handle'))
			tdt = transport_dt(time);
		else
			tdt = transport_dt;
		end
		tdt = min(tdt, 0.5 * idt);
		time_next = time + idt;
		
		% determine typical metabolite concentration scales based on the current and past time points
		concentration_scales = get_scales(0.5*(squeeze(mean(Ccurrent,1))'+squeeze(mean(concentrations(1:(next_tg-1),:,:),[1,2]))));
		
		% determine acceptable concentration errors if needed
		if(any(isnan(concentration_errors)))
			reaction_concentration_errors = reaction_relative_errors .* concentration_scales;
			reaction_concentration_errors(~isnan(concentration_errors)) = concentration_errors(~isnan(concentration_errors));
			transport_concentration_errors = transport_relative_error .* concentration_scales;
			transport_concentration_errors(~isnan(concentration_errors)) = concentration_errors(~isnan(concentration_errors));
		else
			reaction_concentration_errors = concentration_errors;
			transport_concentration_errors = concentration_errors;
		end

		% calculate change due to transport for a short time step, ignoring any reactions, using MATLAB's PDE solver
		parfor (m=1:NM, ternary(parallel,Inf,0))
			PDEsolver_options = odeset('RelTol',transport_relative_error,'AbsTol',transport_concentration_errors(m),'InitialStep',tdt);
			Ccurrent_functional = griddedInterpolant(gridZ, Ccurrent(:,m));
			sol = pdepe(0, @(z,t,C,dCdz) transport_PDE_functional(t,z,C,dCdz,diffusivities{m}(t,z),diffusivity_gradients{m}(t,z),advections{m}(t,z),dilution_slope(z),use_ficks_law),...
						@(z) Ccurrent_functional(z),...
						@(zu,Cu,zd,Cd,t) transport_PDE_BC(t,zu,zd,Cu,Cd,BCtypesUp(m),BCtypesDown(m),BCvaluesUp{m}(t),BCvaluesDown{m}(t),advections{m}(t,zu),advections{m}(t,zd),diffusivities{m}(t,zu),diffusivities{m}(t,zd),diffusivity_gradients{m}(t,zu),diffusivity_gradients{m}(t,zd),use_ficks_law),...
						gridZ, [time, 0.5*(time+time_next), time_next], PDEsolver_options);
			if(size(sol,1)<3)
				errormsg = sprintf('Could not integrate PDE for metabolite %s, between times %g - %g',metabolite_names{m},time,time_next);
				parfor_OK = false;
				continue; % break not allowed in parfor loop
			end
			Ctransported(:,m) = sol(end,:);
		end
		if(not(parfor_OK)) return; end % failure inside parfor loop

		% if some metabolites overshot into the negative, push them up to zero while preserving depth-integrated quantity (i.e., remove some from other depths of positive concentration)
		for m=1:NM
			negatives = interior(find(Ctransported(interior,m)<0));
			if(length(negatives)>0)
				positives 					= interior(find(Ctransported(interior,m)>=0));
				total_negative 				= -sum(Ctransported(negatives,m));
				total_positive 				= sum(Ctransported(positives,m));
				if(diagnostics) fprintf('%s  DIAGNOSTICS: Correcting negative concentrations for transported "%s", at time %g (total neg/total pos = %g)\n',verbose_prefix,metabolite_names{m},time_next/365,total_negative/total_positive); end
				Ctransported(negatives,m) 	= 0;
				Ctransported(positives,m) 	= Ctransported(positives,m) * (1-min(1,total_negative/total_positive)); % deplete the positive regions to correct for the additions in the negative regions. Depletion should be proportional to the current concentration, but at most as much as there is available.
			end
		end
		
		% precalculate equilibrium constants at every depth
		% first calculate logK for every forward reaction, then adopt negated copies for the reversible reactions
		time_column = repmat(time_next,NZ,1);
		if(isa(log_equilibrium_constants,'function_handle'))
			% log_equilibrium_constants is a single function of (t,z) returning NR log-equilibrium constants
			for z=1:NZ
				logK(z,:) = log_equilibrium_constants(time_next,z);
			end
		else
			for r=1:NR
				logK(:,r) = log_equilibrium_constants{r}(time_column,gridZ);
			end
		end
				
		% calculate temperatures (in C) and logK_epsilon at each depth
		temperatures  = temperature(time_column,gridZ);
		logK_epsilons = bsxfun(@rdivide, DeltaG_threshold, ((273.15+temperatures)*Rg));
		
		% simulate fast-reaction step within each grid cell
		if(NR>0)
			Cnext(1,:)	  = Ctransported(1,:);
			Cnext(end,:)  = Ctransported(end,:);
			%fprintf('  Time %g, Ctransported: %s\n  logK: %s',time/365,sprint_named_numerics(metabolite_names, squeeze(Ctransported(50,:))', '%.5g', ', '),sprint_named_numerics(reaction_names, squeeze(logK(50,:))', '%.5g', ', ')); % debug
			parfor (r=1:length(depth_shuffle), ternary(parallel,Inf,0))
				z = depth_shuffle(r);
				%fprintf('  Depth %g..\n    Ctransported: %s\n',gridZ(z),sprint_named_numerics(metabolite_names, squeeze(Ctransported(z,:))', '%.4g', ', ')); % debug
				[Cparfor(r,:), ...
				Rparfor(r,:)] = get_dead_end(Ctransported(z,:)', ...
											logK(z,:)',...
											stoichiometry,...
											reaction_concentration_errors,...
											logK_epsilons(z,:),...
											reaction_relative_errors,...
											reversible,...
											kinetics,...
											reactions_max_runtime);
			end
			Cnext(depth_shuffle,:) = Cparfor;
			reaction_fluxes(depth_shuffle,:) = Rparfor;
		else
			Cnext = Ctransported;
			reaction_fluxes(:) = 0;
		end

		% interpolate concentrations on requested solution grid, if needed
		while((next_tg<=NT) & (time<=gridT(next_tg)) & (time_next>=gridT(next_tg)))
			if(diagnostics) fprintf('%s  DIAGNOSTICS: Storing concentrations and other variables at time %g (#%d out of %d)..\n', verbose_prefix, gridT(next_tg), next_tg, NT); end
			temperatures				  = 273.15+temperatures;
			lambda 						  = (gridT(next_tg)-time)/(time_next-time);
			concentrations(next_tg,:,:)   = lambda*Cnext + (1-lambda)*Ccurrent;
			net_sources(next_tg,:,:) 	  = (Cnext - Ctransported)/idt;
			if(NR>0)
				reaction_rates(next_tg,:,:)	  = reaction_fluxes/idt;
				energy_rates(next_tg,:)		  = -(Rg*temperatures .* sum(Cnext.*log(max(0,Cnext))-Ctransported.*log(max(0,Ctransported)) - (Cnext-Ctransported),2) - Rg*temperatures .* sum(reaction_fluxes.*logK,2));
				for z=1:NZ
					C = squeeze(concentrations(next_tg,z,:));
					deltaLogK(next_tg,z,:) = reaction_deltaLogK(stoichiometry, logK(z,:), C);
					tendencies(next_tg,z,:)= tendencies_logK(stoichiometry, logK(z,:), 2*logK_epsilons(z,:), C);
					%if(z==100) fprintf('storing C=%s.  deltaLogK=%s.  tendencies=%s\n',sprint_numeric_vector(C,', '),sprint_numeric_vector(deltaLogK(next_tg,z,:),', '),sprint_numeric_vector(tendencies(next_tg,z,:),', ')); end % debug
				end
				ndeltaLogK(next_tg,:,:) = deltaLogK(next_tg,:,:)./reshape(abs(logK),[1,NZ,NR]);
			end
			next_tg = next_tg + 1;
		end

		% move to next time point
		time = time_next;
		Ccurrent = Cnext;
	end
end



% returns Gibbs free energy of reactions
% Input:
%   stoichiometry: 2D array of size NM x NR, stoichiometric coefficients of reactions
%   DeltaG0: 1D vector of size NR or 2D array of size NZ x NR, standard Gibbs free energies of reactions. In the former case, DeltaG0 of a reaction is assumed to be the same for all locations.
%	T: 1D vector of size NZ, listing temperatures over NZ locations (in degrees C)
%   concentrations: 2D array of size NZ x NM, listing concentrations of NM metabolites over NZ locations
% Output:
%   DeltaG: 2D array of size NZ x NR, listing Gibbs free energies of reactions
% By convention, DeltaG=Inf whenever a substrate has concentration <=0.
function DeltaG = reaction_DeltaG(stoichiometry, DeltaG0, T, concentrations)
	Rg = 8.3144621e-3; 	% Gas constant (kJ/(K*mol))
	NZ = size(concentrations,1);
	NM = size(concentrations,2);
	NR = size(stoichiometry,2);
	if(size(DeltaG0,1)==1)
		DeltaG0 = DeltaG0'; 
	end
	if(size(DeltaG0,1)==1)
		DeltaG = repmat(reshape(DeltaG0,[1 NR]),NZ,1);
	else
		DeltaG = DeltaG0;
	end
	T = reshape(T,[NZ 1]); % make sure T is column vector
	for r=1:NR
		relevants   = find(stoichiometry(:,r)~=0);
		DeltaG(:,r) = DeltaG(:,r) + Rg * (T+273.15).*(stoichiometry(relevants,r)' * log(concentrations(:,relevants)));

		% set DeltaG of reactions to -Inf if one of the products has zero concentration
		products 		= find(stoichiometry(:,r)>0);
		zero_metabolites = any(concentrations(:,products)<=0, 1);
		DeltaG(zero_metabolites,r) = -inf;

		% set DeltaG of reactions to +Inf if one of the substrates has zero concentration
		% this should come last, so that it overwrites other overlapping scenarios
		substrates 		= find(stoichiometry(:,r)<0);
		zero_metabolites= any(concentrations(:,substrates)<=0, 1);
		DeltaG(zero_metabolites,r) = inf;
	end
end


% determine the direction a reaction would go to based on its logK (log-equilibrium constant) and current concentrations
% this function returns -1 if the reaction would go in reverse, +1 if it would go forward, 0 if it is essentially at equilibrium (up to some tolerance logKepsilon), and NaN if there are substrates & products that are at zero.
function tendencies = tendencies_logK(stoichiometry, logK, logKepsilon, concentrations)
	NR = size(stoichiometry,2);
	tendencies = zeros(NR, 1);
	for r=1:NR
		missing_substrates 	= any(concentrations(stoichiometry(:,r)<0)<=0);
		missing_products 	= any(concentrations(stoichiometry(:,r)>0)<=0);
		if(missing_substrates & missing_products)
			tendencies(r) = NaN;
		elseif(missing_substrates)
			tendencies(r) = -1;
		elseif(missing_products)
			tendencies(r) = 1;
		else
			relevants	= find(stoichiometry(:,r)~=0);
			deltaLogK 	= stoichiometry(relevants,r)' * log(concentrations(relevants)) - logK(r);
			if(deltaLogK >= logKepsilon)
				tendencies(r) = -1;
			elseif(deltaLogK <= -logKepsilon)
				tendencies(r) = 1;
			else
				tendencies(r) = 0;
			end
		end
	end
end



% calculate the difference between the log reaction coefficient and its log equilibrium constant, i.e. deltaLogK:=ln(Q)-ln(K)
% As long as deltaLogK<0 (which is equivalent to DeltaG<0), one would expect the reaction to move forward
% Input:
%   stoichiometry: 2D array of size NM x NR, stoichiometric coefficients of reactions
%   logK: 1D vector of size NR, log-equilibrium constants of reactions
%   concentrations: 1D column vector of size NM, listing concentrations of NM metabolites
% Output:
%   deltaLogK: 1D column vector of size NR, listing the deltaLogK of each reaction
% By convention, deltaLogK=Inf whenever a substrate has concentration <=0.
function deltaLogK = reaction_deltaLogK(stoichiometry, logK, concentrations)
	NR = size(stoichiometry,2);
	deltaLogK = zeros(NR, 1);
	for r=1:NR
		relevants = find(stoichiometry(:,r)~=0);

		deltaLogK(r) = stoichiometry(relevants,r)' * log(concentrations(relevants)) - logK(r);

		% set deltaLogK of reactions to -Inf if one of the products has zero concentration
		if(any(concentrations(stoichiometry(:,r)>0)<=0))
			deltaLogK(r) = -inf;
		end

		% set deltaLogK of reactions to +Inf if one of the substrates has zero concentration
		% this should come last, so that it overwrites other overlapping scenarios
		if(any(concentrations(stoichiometry(:,r)<0)<=0))
			deltaLogK(r) = inf;
		end
	end
end



% Calculate the normalized deltaLogK, i.e., deltaLogK./abs(logK), and optionally its Jacobian w.r.t. Hmod, where Htrue = Hmod*scales + shifts
% stoichiometry must be an NM x NR matrix
% logK, Hmod, Hscales and Hshifts must each be a column vector of size NR
% Cstart must be a column vector of size NM
function [deltaLogK,deltaLogK_Jacobian] = normalized_deltaLogK(stoichiometry, logK, Cstart, Hmod, Hscales, Hshifts)
	NR = size(stoichiometry,2);
	deltaLogK = zeros(NR, 1);
	if(nargout>1) deltaLogK_Jacobian = zeros([NR,NR]); end
	C = Cstart + stoichiometry * (Hmod.*Hscales + Hshifts);
		
	for r=1:NR
		relevants = find(stoichiometry(:,r)~=0);

		deltaLogK(r) = stoichiometry(relevants,r)' * log(C(relevants)) - logK(r);
		if(nargout>1) 
			deltaLogK_Jacobian(r,:) = Hscales.*stoichiometry(relevants,:)' * (stoichiometry(relevants,r)./C(relevants));
		end

		% set deltaLogK of reactions to -Inf if one of the products has zero concentration
		if(any(C(stoichiometry(:,r)>0)<=0))
			deltaLogK(r) = -inf;
		end

		% set deltaLogK of reactions to +Inf if one of the substrates has zero concentration
		% this should come last, so that it overwrites other overlapping scenarios
		if(any(C(stoichiometry(:,r)<0)<=0))
			deltaLogK(r) = inf;
		end
	end
	% account for rescalings of the returned function
	deltaLogK = deltaLogK./abs(logK);
	if(nargout>1)
		deltaLogK_Jacobian = bsxfun(@rdivide, deltaLogK_Jacobian, abs(logK));
	end
end

function [SSdeltaLogK, SSJacobian] = SS_normalized_deltaLogK(stoichiometry, logK, Cstart, Hmod, Hscales, Hshifts)
	[deltaLogK,deltaLogK_Jacobian] = normalized_deltaLogK(stoichiometry, logK, Cstart, Hmod, Hscales, Hshifts);
	SSdeltaLogK = sum(deltaLogK.^2);
	if(nargout>1)
		SSJacobian = 2*sum(bsxfun(@times,deltaLogK,deltaLogK_Jacobian),1);
	end
end


% Find the thermodynamic equilibrium of a set of reversible reactions, by jointly solving deltaLogK = 0 for all reactions by choice of the reaction fluxes H (where C = Cstart + stoichiometry*H)
% Inputs:
%	stoichiometry: NM x NR stoichiometric matrix of all NR reactions
%	logK: Column vector of size NR, specifying the logarithms of the modified equilibrium constants, i.e., such that a reaction is at equilibrium when sum_m stoichiometry(m,r) * log(C_m) is equal to logK
%	Cstart: Column vector of size NM, specifying the initial metabolite concentrations
%	logEpsilon: Column vector of size NR of non-negative numbers, zero-threshold for deltaLogK, for considering a reaction at equilibrium ("dead"). This should encode the numerical error tolerance.
%	max_runtime: Optional maximum number of seconds (wall time) to allocate for the computation, before giving up. If NaN, no time limit is imposed.
%
% Explanation of the algorithm:
% This function searches for the thermodynamic equilibrium C and associated reaction flux H, i.e. solve the system of equations deltaLogK(C=Cstart+stoichiometry*H)=0 for the unknown reaction flux vector H
% Due to the sudden steepness of the gradient near the (a priori unknown) equilibrium, standard solvers or least-squares minimizers can easily get stuck somewhere near the equilibrium (in terms of C & H) but still far from equilibrium (in terms of deltaLogK).
% The following tricks have led to success after many days of troubleshooting:
%	- Use fsolve instead of simply fminunc or lsqnonlin
%	- Normalize the reaction scales (i.e., |H|) as well as the function scales (by dividing deltaLogK by abs(logK)) during fitting/solving
%	- Do provide the gradient (Jacobian) of the functions to be zeroed (i.e., of deltaLogK/abs(logK))
%	- Repeat fsolve multiple times, each time using the previously found C as a new starting point, and solving for the H that would get you closer to equilibrium.
%	  That is, repeatedly solve for H_new such that C_previous + stoichiometry * H_new comes as close as possible to equilibrium, while also updating the normalization (typical scales) of H_new as we move closer to equilibrium.
%	- With each fsolve repeat, choose the scalings for H based on the Jacobian at the starting point, such that the Jacobian w.r.t. the scaled H would have modulus 1 on average. This dramatically improves the ability to find the equilibrium.
%	- The 'levenberg-marquardt' algorithm and the default fsolve algorithm ('trust-region dogleg') yield similarly good results, but the default is much faster.
%	- Common non-linear transformations of the objective function (e.g. log or exp or sigmoid) did not improve results.
%	  The issue is that the location of the steep section of the objective functions is at the equilibrium, which is a priori unknown, so it's impossible to come up with an a priori transformation that would "dampen" that steepness.
%	- Repeat the fitting with multiple alternative starting points. Even for simple systems with 2 reactions, sometimes >1000 trials are needed to get a good starting point and find the equilibrium.
function [Ceq, Heq, warning] = find_equilibrium(stoichiometry, logK, Cstart, logKepsilon, max_runtime)
	start_tic		= tic;
	NR				= size(stoichiometry,2);
	Hzero			= zeros(NR,1);
	Nrounds 		= 100; % maximum number of fsolve rounds per fitting trial
	Ntrials			= 100;
	nlogKepsilon	= logKepsilon./abs(logK);
	fitoptions		= optimoptions('fsolve','Display', 'off', 'SpecifyObjectiveGradient',true, 'StepTolerance', 1e-8, 'FunctionTolerance', 1e-8, 'MaxIterations', NR*100, 'MaxFunctionEvaluations', NR*1000, 'OptimalityTolerance', 1e-8);
	start_Hscales 	= get_scales(abs(pinv(stoichiometry) * Cstart));

	trial 	 		= 1;
	best_trial		= 0;
	best_Ceq 		= Cstart;
	best_Heq 		= Hzero;
	best_ndeltaLogK = reaction_deltaLogK(stoichiometry, logK, best_Ceq)./abs(logK);
	best_Nrounds	= 0;
	abort			= false;
	%fprintf('  finding equilibrium:\n    start_Hscales=%s\n    logK=%s\n    nlogKepsilon=%s\n    logKepsilon=%s\n',sprint_numeric_vector(start_Hscales,', '),sprint_numeric_vector(logK,', '),sprint_numeric_vector(nlogKepsilon,', '),sprint_numeric_vector(logKepsilon,', ')); % debug
	while(trial<=Ntrials)
		% check if we should abort due to time out
		if((~isnan(max_runtime)) && (toc(start_tic)>=max_runtime)) break; end
	
		% determine starting point for fsolve
		if(trial==1)
			Heq = Hzero;
		elseif(mod(trial,3)==0)
			Heq = start_Hscales .* sign(2*rand(NR,1)-1) .* (10.^(5*rand(NR,1)-5));
		elseif(mod(trial,3)==1)
			Heq = best_Heq .* sign(2*rand(NR,1)-1) .* (10.^(8*rand(NR,1)-4));
		else
			Heq = best_Heq .* (4*rand(NR,1)-2);
		end
		Ceq = Cstart + stoichiometry * Heq;
		if(any(Ceq<=0)) continue; end

		% choose Hscales based on Jacobian at starting point
		[ndeltaLogK,J] 	= normalized_deltaLogK(stoichiometry, logK, Ceq, Hzero, start_Hscales, Hzero);
		Hscales = get_scales(start_Hscales ./ (mean(abs(J),1)'));
	
		for round=1:Nrounds
			% check if we should abort due to time out
			if((~isnan(max_runtime)) && (toc(start_tic)>=max_runtime))
				abort = true;
				break;
			end
	
			[Hadded,ndeltaLogK,exitflag,details] = fsolve(@(H) normalized_deltaLogK(stoichiometry, logK, Ceq, H, Hscales, Hzero), Hzero, fitoptions);
			if(all(Hadded==0))
				if(any(Heq==0))
					% in some situations (particularly when Cstart is close to zero) fsolve gets stuck on the initial point, 
					% due to Hscales being severely overestimatedhus overshooting into non-physical Ceq. So try with reduced Hscales.
					Hscales(Heq==0) = 0.01 * Hscales(Heq==0);
				else
					% we did move in a previous iteration, just not on this last one. So no need to keep trying, since it seems that we converged.
					break;
				end
			else
				%fprintf('    n=%d.  moving a bit. Hadded=%s\n',n,sprint_numeric_vector(Hadded,', ')); % debug
				Hadded 	= Hadded .* Hscales;
				Ceq		= Ceq + stoichiometry * Hadded; % update Ceq, moving closer to equilibrium
				Heq		= Heq + Hadded; % add the new reaction flux increment to the total reaction flux
				if(all(abs(ndeltaLogK)<nlogKepsilon)) break; end

				% update scales at new point based on local Jacobian
				[~,J] = normalized_deltaLogK(stoichiometry, logK, Ceq, Hzero, Hscales, Hzero);
				Hscales = get_scales(Hscales ./ (mean(abs(J),1)'));
			end
		end
		if(max(abs(ndeltaLogK./nlogKepsilon))<max(abs(best_ndeltaLogK./nlogKepsilon)))
			% this Ceq is the best (closest to equilibrium) found so far, so keep a record of it
			best_ndeltaLogK	= ndeltaLogK;
			best_Ceq 		= Ceq;
			best_Heq 		= Heq;
			best_trial		= trial;
			best_Nrounds	= round;
		end
		if(all(abs(ndeltaLogK)<nlogKepsilon)) break; end
		if(abort) break; end
		trial = trial + 1;
	end
	Ceq = best_Ceq;
	Heq = best_Heq;
end


% given some start concentrations Cstart and equilibrium constants K for a set of possible reactions, find the final (dead-end) concentrations Cfinal, i.e. after all reactions have "completed", under constraint that a reaction only ever proceeds if its DeltaG is strictly negative (or its reaction coefficient is strictly smaller than its equilibrium constant).
function [Cfinal,...			% 1D column vector of size NM, listing the final metabolite concentrations
		  Hfinal...				% 1D column vector of size NR, listing the cumulative flux through each reaction
		  ] = get_dead_end(	Cstart,...			% 1D numeric column vector of size NM
							logK,...			% 1D numeric vector of size NR
							stoichiometry,...	% 2D numeric matrix of size NM x NR
							concentration_errors,...
							logKepsilon,...		% column vector of size NR of non-negative numbers, zero-threshold for deltaLogK, for considering a reaction complete ("dead"). For irreversible (i.e., biological) reactions this should encode the viability threshold (energy quantum), whereas for reversible (i.e., abiotic) reactions this should encode the numerical error tolerance.
							relative_errors,... % scalar or numeric column vector of size NM+NR, allowed relative error(s) for ODE solver. Typical values are ~1e-6
							reversible,...		% column vector of size NR, specifying which reactions are reversible
							kinetics,...		% optional function handle that takes C (concentrations) as argument and returns a vector of NR reaction rates. Only the relative rates matter, since the ODE is integrated until reaching a dead-end. If NaN, then reaction rates are proportional to deltaLogK. Kinetics are only supported if all reactions are non-reversible.
							max_runtime)		% optional maximum number of seconds (wall time) to allocate for the computation (e.g., the ODE solver), before giving up. This may be needed to avoid the solvers getting stuck. Set this to NaN to not impose any time limit.
	NM = length(Cstart);
	NR = length(logK);
	if(NR==0)
		% no reactions to run
		Cfinal = Cstart;
		Hfinal = zeros(0,1);
		return;
	end
	logK 			= reshape(logK, [NR 1]); % make sure logK is a column vector
	logKepsilon 	= reshape(logKepsilon, [NR 1]);
	has_reversibles	= any(reversible);
	reversibles		= find(reversible);
	irreversibles	= find(~reversible);
	use_kinetics 	= (isa(kinetics, 'function_handle') & (~has_reversibles));
	start_tic 		= tic;

	Cstart(Cstart<=0) = 1e-20 * concentration_errors(Cstart<=0);
	reaction_scales = get_scales(abs(pinv(stoichiometry) * Cstart));
	
	% determine rate factors and max reaction rate for ODE solver
	% only relevant if none of the reactions are reversible
	if(use_kinetics)
		rate_factors	= repmat(mean(reaction_scales) / mean(abs(kinetics(Cstart))), [NR,1]);
		max_rate 		= 100*max(reaction_scales);
	else
		rate_factors	= mean(reaction_scales) ./ abs(logK);
		max_rate 	 	= 100*max(reaction_scales); % maximum allowed reaction rate
	end
	
	% specify an "event" function that tells the ODE solver to terminate when a dead-end is reached, based on the current metabolite concentrations C
	% the value of this function should be negative whenever the solver should continue, and become positive when the solver should terminate
	function [value,isterminal,direction] = dead_end_event(C)
		runtime = toc(start_tic);
		if((~isnan(max_runtime)) && (runtime>=max_runtime))
			% computation timed out
			isterminal 	= 1;
			direction	= 0;
			value		= 1;
			return;
		end
		if(has_reversibles)
			deltaLogK 	= reaction_deltaLogK(stoichiometry(:,irreversibles), logK(irreversibles), C);
			value 		= max(-1,min(1,min((logKepsilon(irreversibles)+deltaLogK)./abs(logK(irreversibles)))));
		else
			deltaLogK 	= reaction_deltaLogK(stoichiometry, logK, C);
			value 		= max(-1,min(1,min((logKepsilon+deltaLogK)./abs(logK))));
		end
		isterminal 	= 1;
		direction  	= 1;
	end
	
	% define vector field for ODE solver, in the case where all reactions are non-reversible (i.e., no strict equilibrium conditions need to be satisfied)
	% state variables for the ODE will be the concentrations C_m(t) & cumulative reaction fluxes, H_r(t) = \int_0^t rate_r(s) ds
	% by considering both the concentrations and the cumulative reaction fluxes as state variables, we can keep track of the final concentrations as well as the final cumulative reaction fluxes (needed e.g. for energy calculations)
	% Optionally, this function also returns the Jacobian of the vector field (only supported if use_kinetics=false)
	function [dCHdt,Jacobian] = vector_field(t,CH)
		C = CH(1:NM); % extract concentrations from state variable
		deltaLogKs = reaction_deltaLogK(stoichiometry, logK, C); % calculate deviation from equilibrium for each reaction
		if(use_kinetics)
			rates = rate_factors .* kinetics(C);
		else
			rates = - rate_factors .* deltaLogKs; % set reaction rates proportional to deviation from equilibrium
		end
		rates((deltaLogKs+logKepsilon>=0) & (~reversible)) = 0; % prevent endergonic non-reversible reactions from running
		rates((rates<0) & (~reversible)) = 0; % prevent non-reversible reactions from running backwards
		rates = max(-max_rate,min(max_rate,rates)); % restrict non-reversible reaction rates to [-max_rate, max_rate]
		dCHdt = [stoichiometry*rates; rates];

		if((nargout>1) & (~use_kinetics))
			% calculate Jacobian ddCHdt/dCH
			drates_dC = zeros(NR,NM);
			for r=1:NR
				relevants = find(stoichiometry(:,r)~=0);
				drates_dC(r,relevants) = -rate_factors(r) * stoichiometry(relevants,r)./C(relevants);
			end
			drates_dH = drates_dC * stoichiometry;
			Jacobian  = zeros(NR+NM,NR+NM);
			Jacobian(1:NM,1:NM) 					= stoichiometry * drates_dC;
			Jacobian(1:NM,(NM+1):(NM+NR)) 			= stoichiometry * drates_dH;
			Jacobian((NM+1):(NM+NR),1:NM) 			= drates_dC;
			Jacobian((NM+1):(NM+NR),(NM+1):(NM+NR)) = drates_dH;
		end
	end

	% calculate vector field for DAE system, in the case where some reactions are reversible and others are non-reversible
	% the state variables are the rescaled cumulative reaction fluxes H, such that C(t) = Cstart + stoichiometry * (H.*Hscales)
	Hzero = zeros(NR,1);
	function [f, dfdH] = DAE_model(H, Hscales)
		C  = Cstart + stoichiometry * (H.*Hscales); % calculate concentrations from state variable H

		% get the normalized deltaLogK (normalized over abs(logK)) and its associated Jacobian w.r.t. H
		if(nargout>1)
			[ndeltaLogK, ndeltaLogK_Jacobian] = normalized_deltaLogK(stoichiometry, logK, Cstart, H, Hscales, Hzero);
		else
			ndeltaLogK = normalized_deltaLogK(stoichiometry, logK, Cstart, H, Hscales, Hzero);
		end
		
		f  				 = zeros(NR,1);
		f(reversibles)   = ndeltaLogK(reversibles);
		f(irreversibles) = - (ndeltaLogK(irreversibles) + logKepsilon(irreversibles) ./ abs(logK(irreversibles)));
		if(nargout>1)
			% use chain rule of derivatives to calculate Jacobian df/dH = df/dndeltaLogK * ndeltaLogK_Jacobian
			dfdl = ones(NR,1);
			dfdl(irreversibles) = -1;
			dfdH = bsxfun(@times,dfdl,ndeltaLogK_Jacobian);
		end		
	end

	if(all(reversible))
		% all reactions are reversible, so solve a system of algebraic equations for local thermodynamic equilibrium
		[Cfinal, Hfinal] = find_equilibrium(stoichiometry, logK, Cstart, logKepsilon, max_runtime);

	elseif(has_reversibles)
		% use DAE solver, treat integrated reaction fluxes as the sole state variables
		% reversible reactions are assumed to be at thermodynamic equilibrium at all times
		Hstart = zeros(NR,1);
		[Ceq, Hstart(reversibles)] = find_equilibrium(stoichiometry(:,reversibles), logK(reversibles), Cstart, min(logKepsilon(reversibles),1e-6 * abs(logK(reversibles))), max_runtime);
		
		% update scales of reversible (equilibrated) reactions
		Hscales = reaction_scales;
		Hscales(reversibles) = get_scales(Hstart(reversibles));
		
		% check if we already reached a dead end
		[dead_end_status, ~, ~] = dead_end_event(Cstart+stoichiometry*Hstart);
		if(dead_end_status>=0)
			% Cstart is already a dead-end
			Cfinal = Cstart;
			Hfinal = Hstart;
			return;
		end
		
		% solve DAE system, with algebraic equations imposing thermodynamic equilibrium for reversible reactions and differential equations specifying the reaction rates for non-reversible reactions.
		% The state variables are the integrated reaction fluxes H
		% It is important that the initial condition for this DAE system already satisfies the algebraic equations, i.e., where reversible reactions are at equilibrium.
		% Tests have also shown that the solver works much better when provided with the Jacobian of the DAE_model
		max_time = 100;
		M = diag(double(~reversible)); % repare mass matrix for DAE
		warning('off','MATLAB:ode15s:IntegrationTolNotMet');
		[~,Hfinal,event_t,event_H, ~] = ode15s(@(t,H) DAE_model(H,Hscales),...
												[0, max_time],...
												Hstart./Hscales,...
												odeset('Mass',M,'MassSingular','yes','MStateDependence','none','Jacobian',@(t,H) second_part(@DAE_model, H, Hscales), 'RelTol',relative_errors,'AbsTol',relative_errors,'Events',@(t,H) dead_end_event(Cstart+stoichiometry*(H.*Hscales))));
		if(length(event_H)>0)
			Hfinal = Hscales .* event_H(end,:)';
		else
			Hfinal = Hscales .* Hfinal(end,:)';
		end
		Cfinal = Cstart + stoichiometry*Hfinal;
		
		%fprintf('  deltaLogK/|logK|=%s\n',sprint_numeric_vector(reaction_deltaLogK(stoichiometry, logK, Cfinal)./abs(logK),', ')); % debug

	else
		% none of the reactions are reversible, so use classical ODE solver
		Hstart  = zeros(NR,1);
		CHstart = [Cstart; Hstart];
		[dead_end_status, ~, ~] = dead_end_event(Cstart);
		if(dead_end_status>=0)
			% Cstart is already a dead-end
			Cfinal = Cstart;
			Hfinal = Hstart;
			return;
		end
		max_time = 100;
		[~,CHfinal,event_t,event_CH, ~] = ode45(@vector_field, [0, max_time], CHstart,...
												odeset('RelTol',relative_errors,'AbsTol',[concentration_errors; relative_errors*reaction_scales],'Events',@(t,CH) dead_end_event(CH(1:NM)),'NonNegative',1:NM));
		if(length(event_CH)>0)
			CHfinal = event_CH(end,:)';
		else
			CHfinal = CHfinal(end,:)';
		end
		Cfinal = CHfinal(1:NM);
		Hfinal = CHfinal((NM+1):(NM+NR));
	end	
end


function s = sprint_numeric_vector(values,sep)
	s = strjoin(cellfun(@num2str,num2cell(values),'uniformoutput',false),sep);
end


% Ternary operator for C-like conditional expressions
function c = ternary(condition, result_true, result_false)
    if condition
        c = result_true;
    else
        c = result_false;
    end
end


function scales = get_scales(X)
	scales 		= abs(X);
	valids 		= find(((scales~=0) & (~isnan(scales)) & (~isinf(scales))));
	invalids 	= find(((scales==0) | isnan(scales) | isinf(scales)));
	if(length(valids)==0)
		scales = ones(size(X));
	elseif(length(invalids)>0)
		scales(invalids) = mean(scales(valids));
	end
end



% type_u and type_d can be 0 (Dirichlet, given value) or 1 (Robin, given influx) or -1 (Neumann, given gradient) or -2 (given relative gradient)
% Cu & Cd are concentrations at the top & bottom boundaries
% vu & vd are advection velocities at the top & bottom boundaries (positive means downwards)
% dKudz & dKddz are derivatives of the diffusion coefficient (dD/dz) at the top & bottom boundaries. Only relevant if use_ficks_law==false.
function [pu,qu,pd,qd] = transport_PDE_BC(t,zu,zd,Cu,Cd,type_u,type_d,value_u,value_d,vu,vd,Ku,Kd,Kudz,Kddz,use_ficks_law)
	pu = NaN; qu = NaN; pd = NaN; qd = NaN; % default return values
	if(isnan(value_u)) fprintf('ERROR: Could not solve PDE: Upper BC is NAN at t=%g\n',t); return; end
	if(isnan(value_d)) fprintf('ERROR: Could not solve PDE: Bottom BC is NAN at t=%g\n',t); return; end
	if(type_u==0)
		% Dirichlet
		pu=Cu-value_u; qu=0;
	elseif(type_u==1)
		% Robin (influx)
		pu=value_u; qu=1;
	elseif(type_u==-1)
		% Neumann (gradient)
		if(use_ficks_law)
			pu=vu*Cu-value_u*Ku; qu=1;
		else
			pu=vu*Cu-value_u*Ku-Cu*Kudz; qu=1;
		end
	else
		% Relative gradient
		if(use_ficks_law)
			pu=vu*Cu-value_u*Cu*Ku; qu=1;
		else
			pu=vu*Cu-value_u*Cu*Ku-Cu*Kudz; qu=1;
		end
	end
	
	if(type_d==0)
		% Dirichlet
		pd=Cd-value_d; qd=0;
	elseif(type_d==1)
		% Robin (influx)
		pd=value_d; qd=-1;
	elseif(type_d==-1)
		% Neumann (gradient)
		if(use_ficks_law)
			pd=vd*Cd-value_d*Kd; qd=1;
		else
			pd=vd*Cd - value_d*Kd - Cd*Kddz; qd=1;
		end
	else
		% Relative gradient
		if(use_ficks_law)
			pd=vd*Cd-value_d*Cd*Kd; qd=1;
		else
			pd=vd*Cd-value_d*Cd*Kd-Cd*Kddz; qd=1;
		end
	end
end





function [c,f,s] = transport_PDE_functional(t,z,C,dCdz,D,dDdz,v,rdAdz,use_ficks_law)
	f = NaN; s = NaN; % default return values
	c = 1; % time-derivative coefficient
	if(isnan(D))
		fprintf('ERROR: Could not solve PDE: Diffusion coefficient is NaN at t=%g, z=%g\n',t,z); return;
	elseif(isnan(v))
		fprintf('ERROR: Could not solve PDE: Advection coefficient is NaN at t=%g, z=%g\n',t,z); return;
	end
	if(use_ficks_law)
		f = D*dCdz - v*C; % flux term = Fick's law = D*p_zC
	else
		f = D*dCdz + dDdz*C - v*C; % flux term = p_z(D*C)
	end
	s = rdAdz .* f; % source-like term accounting for geometric dilution
end


function summary = sprint_named_numerics(names, values, number_formatting, delimiter)
	summary = strjoin(strcat(names,'=',arrayfun(@(x) num2str(x,number_formatting),values,'Uniform',false)),delimiter);
end


% print a 1D or 2D matrix to screen
% formatting can be e.g. '%.g', sep can be e.g. '  '.
function print_matrix(M, linePrefix, sep, formatting)
	if((size(M,1)<2) || (size(M,2)<2)) % matrix is 1-dimensional. treat as row-vector
		format = [formatting repmat([sep formatting],1,length(M)-1)];
	else
		lformat = [formatting repmat([sep formatting],1,size(M,2)-1)];
		format = [repmat([linePrefix lformat '\n'],1,size(M,1))];
	end
	fprintf(format, M);
end



function SA = slice(A, indices)
	SA = A(indices);
end


% execute a function that returns 2 values, and extract only the 2nd value
% this may be used e.g. in anonymous functions that make use of the 2nd return value of another function
function B = second_part(fun, varargin)
	[~,B] = fun(varargin{:});
end

